"""
Helper functions for producing plots on simplexes
"""
from typing import Tuple, List, Callable, Optional, Union

import numpy as np
from scipy.optimize import root
from .. import (calculate_nb_states, sample_simplex, )


def simplex_iterator(scale: int, boundary: bool = True) -> Tuple[int, int, int]:
    """
    Systematically iterates through a lattice of points on the 2-simplex.

    Parameters
    ----------
    scale: int
        The normalized scale of the simplex, i.e. N such that points (x,y,z)
        satisify x + y + z == N
    boundary: bool
        Include the boundary points (tuples where at least one
        coordinate is zero)
    Yields
    ------
    Tuple[int, int, int]
        3-tuples, There are binom(n+2, 2) points (the triangular
        number for scale + 1, less 3*(scale+1) if boundary=False

    Citing
    ------
    This function has been copied from: https://github.com/marcharper/python-ternary/blob/master/ternary/helpers.py
    """
    start = 0
    if not boundary:
        start = 1
    for i in range(start, scale + (1 - start)):
        for j in range(start, scale + (1 - start) - i):
            k = scale - i - j
            yield i, j, k


def xy_to_barycentric_coordinates(x: Union[float, np.ndarray], y: Union[float, np.ndarray],
                                  corners: np.ndarray) -> np.ndarray:
    """
    Transforms cartesian into barycentric coordinates.

    Parameters
    ----------
    x : Union[float, numpy.ndarray]
        first component of the cartesian coordinates
    y : Union[float, numpy.ndarray]
        second component of the cartesian coordinates
    corners : numpy.ndarray
        a list or a vector containing the coordinates of the corners

    Returns
    -------
    numpy.ndarray
        The transformmation of the coordinates into barycentric.

    Examples
    --------
    >>> from egttools.plotting import Simplex2D
    >>> simplex = Simplex2D()
    >>> cartesian_coords = np.array([0.2, 0.])
    >>> xy_to_barycentric_coordinates(cartesian_coords[0], cartesian_coords[1], simplex.corners)
    array([0.2, 0. ])
    """
    corner_x = corners.T[0]
    corner_y = corners.T[1]
    x_1 = corner_x[0]
    x_2 = corner_x[1]
    x_3 = corner_x[2]
    y_1 = corner_y[0]
    y_2 = corner_y[1]
    y_3 = corner_y[2]
    l1 = ((y_2 - y_3) * (x - x_3) + (x_3 - x_2) * (y - y_3)) / (
            (y_2 - y_3) * (x_1 - x_3) + (x_3 - x_2) * (y_1 - y_3))
    l2 = ((y_3 - y_1) * (x - x_3) + (x_1 - x_3) * (y - y_3)) / (
            (y_2 - y_3) * (x_1 - x_3) + (x_3 - x_2) * (y_1 - y_3))
    l3 = 1 - l1 - l2
    return np.asarray([l1, l2, l3])


def barycentric_to_xy_coordinates(point_barycentric: np.ndarray, corners: np.ndarray) -> np.ndarray:
    """
    Transforms barycentric into cartesian coordinates.

    Parameters
    ----------
    point_barycentric: numpy.ndarray
        An array containing the 3 barycentric coordinates.
    corners: numpy.ndarray
        An matrix containing the cartesian coordinates of the corners of the triangle that represents the 2-simplex.

    Returns
    -------
    numpy.ndarray
        An array containing the cartesian coordinates of the input point.
    """
    return (corners.T @ point_barycentric.T).T


def calculate_stability(roots: List[np.ndarray], f: Callable[[np.ndarray], np.ndarray]) -> List[bool]:
    """
    Calculates the stability of the roots. It will return a list indicating whether each root
    is or not stable.

    Parameters
    ----------
    roots: numpy.ndarray
        A list or arrays which contain the barycentric coordinates of the roots.
    f: Callable[[numpy.ndarray], numpy.ndarray]
        A function which computes the gradient at any point in the simplex.

    Returns
    -------
    List[bool]
        A list of booleans indicating whether each root is or not stable.
    """
    stability = []
    for stationary_point in roots:
        stable = False
        # first we check if the root is in one of the edges
        if np.isclose(stationary_point, 0., atol=1e-7).any():
            sign_plus = 0
            sign_minus = 0
            # Check if we are in an edge
            edge = np.where(~np.isclose(stationary_point, 0., atol=1e-7))[0]
            if edge.shape[0] > 1:  # we are at an edge
                # Now we perturb the edge
                if stationary_point[edge[0]] + 0.1 <= 1.:
                    tmp = stationary_point.copy()
                    tmp[edge[0]] += 0.1
                    tmp[edge[1]] -= 0.1
                    grad = f(tmp)
                    if grad[edge[0]] > 0:
                        sign_plus = 1
                if stationary_point[edge[0]] - 0.1 >= 0.:
                    tmp = stationary_point.copy()
                    tmp[edge[0]] -= 0.1
                    tmp[edge[1]] += 0.1
                    grad = f(tmp)
                    if grad[edge[0]] > 0:
                        sign_minus = 1
                if sign_minus and not sign_plus:
                    stable = True
            else:  # we are at a vertex
                # Now we perturb the vertex
                tmp = stationary_point.copy()
                tmp[edge[0]] -= 0.1
                tmp[(edge[0] + 1) % 3] += 0.1
                grad = f(tmp)
                if grad[edge[0]] > 0.:
                    sign_plus = 1
                tmp[(edge[0] + 1) % 3] -= 0.1
                tmp[(edge[0] + 2) % 3] += 0.1
                grad = f(tmp)
                if grad[edge[0]] > 0.:
                    sign_minus = 1

                if sign_plus and sign_minus:
                    stable = True
        else:  # we are in the interior of the simples
            # here to analyse stability we need to
            # check wether change in any direction leads back to the point
            unstable = False
            for vertex in range(stationary_point.shape[0]):
                tmp = stationary_point.copy()
                if stationary_point[vertex] + 0.1 <= 1.:
                    tmp[vertex] += 0.1
                    if tmp[(vertex + 1) % 3] - 0.1 >= 0.:
                        tmp[(vertex + 1) % 3] -= 0.1
                        grad = f(tmp)
                        if grad[vertex] > 0.:
                            unstable = True
                            break
                        tmp[(vertex + 1) % 3] += 0.1
                    if tmp[(vertex + 2) % 3] - 0.1 >= 0.:
                        tmp[(vertex + 1) % 3] -= 0.1
                        grad = f(tmp)
                        if grad[vertex] > 0.:
                            unstable = True
                            break
                        tmp[(vertex + 1) % 3] += 0.1
                    tmp[vertex] -= 0.1
                if stationary_point[vertex] - 0.1 >= 0.:
                    tmp[vertex] -= 0.1
                    if tmp[(vertex + 1) % 3] + 0.1 <= 1.:
                        tmp[(vertex + 1) % 3] += 0.1
                        grad = f(tmp)
                        if grad[vertex] < 0.:
                            unstable = True
                            break
                        tmp[(vertex + 1) % 3] += 0.1
                    if tmp[(vertex + 2) % 3] + 0.1 <= 1.:
                        tmp[(vertex + 1) % 3] += 0.1
                        grad = f(tmp)
                        if grad[vertex] < 0.:
                            unstable = True
                            break
            stable = not unstable
        stability.append(stable)
    return stability


def calculate_stationary_points(x: np.ndarray, y: np.ndarray, corners: np.ndarray,
                                f: Callable[[np.ndarray], np.ndarray],
                                border: Optional[int] = 5,
                                delta: Optional[float] = 1e-12,
                                atol: Optional[float] = 1e-7) -> Tuple[List[np.ndarray], List[np.ndarray]]:
    """
    Finds the roots of f (the points where the gradient is 0), given a number of points.

    Parameters
    ----------
    x: numpy.ndarray
        x (cartesian) coordinates of the points for which to look for the gradients.
    y: numpy.ndarray
        y (cartesian) coordinates of the points for which to look for the gradients.
    corners: numpy.ndarray
        A matrix containing the cartesian coordinates of the vertices of the triangle that forms the 2-simplex.
    f: Callable[[numpy.ndarray], numpy.ndarray]
        A function that calculates the gradient at any point in the simplex.
    border: int
        Indicates how close to the simplex borders should we look for the gradients. This allows to avoid
        boundary problems.
    delta: float
        tolerance for considering points outside the simplex.
    atol: float
        tolerance for considering that two roots are equal.

    Returns
    -------
    Tuple[List[numpy.ndarray], List[numpy.ndarray]]
        A list with the barycentric coordinates of all the roots that were found and another list with
        the cartesian coordinates.
    """
    roots = []
    for x, y in zip(x[border:-border], y[border:-border]):
        start = xy_to_barycentric_coordinates(x, y, corners)
        sol = root(f, start, method="hybr")  # ,xtol=1.49012e-10,maxfev=1000
        if sol.success:
            v = sol.x
            if check_if_point_in_unit_simplex(v, delta):
                # only add new fixed points to list
                if not np.array([np.allclose(v, x, atol=atol) for x in roots]).any():
                    roots.append(v)

    return roots, [barycentric_to_xy_coordinates(x, corners) for x in roots]


def find_roots_in_discrete_barycentric_coordinates(f: Callable[[np.ndarray], np.ndarray],
                                                   simplex_size: int,
                                                   nb_edge_points: Optional[int] = None,
                                                   nb_interior_points: Optional[int] = 1000,
                                                   delta: Optional[float] = 1e-12,
                                                   atol: Optional[float] = 1e-3) -> List[np.ndarray]:
    """
    Searches for the roots inside the simplex and returns them in barycentric coordinates.

    Parameters
    ----------
    f: Callable[[numpy.ndarray], numpy.ndarray]
        A function that calculates the gradient of any point inside the simplex.
    simplex_size : int
        Discrete size of the edges of the simplex. This should correspond to the size of the finite population
        in Moran dynamics.
    nb_edge_points: int
        Can be used to explore more points than the existing simplex size.
    nb_interior_points: int
        Number of points to explore inside the simplex.
    delta: float
        Tolerance to consider a point outside the unit simplex.
    atol: float
        Tolerance to consider two roots to be equal.

    Returns
    -------
    List[numpy.ndarray]
        A list with the barycentric coordinates of the roots.

    See Also
    --------
    egttools.plotting.helpers.calculate_stationary_points
    """
    roots = []

    if nb_edge_points is None:
        nb_edge_points = simplex_size

    # We first test values along the edges
    values = np.linspace(0, simplex_size, nb_edge_points, )
    point = np.zeros(shape=(3,), dtype=np.int64)
    for value in values:
        for i in range(3):
            point[i] = value
            point[(i + 1) % 3] = simplex_size - value
            point[(i + 2) % 3] = 0

            sol = root(f, point, method="hybr")  # ,xtol=1.49012e-10,maxfev=1000
            if sol.success:
                v = sol.x / simplex_size
                if check_if_point_in_unit_simplex(v, delta):
                    # only add new fixed points to list
                    if not np.array([np.allclose(v, x, atol=atol) for x in roots]).any():
                        roots.append(v)

    # finally we can explore values inside the simplex
    nb_states = calculate_nb_states(simplex_size, 3)

    if nb_interior_points > nb_states:
        nb_interior_points = nb_states
    initial_points = np.random.choice(range(nb_states), size=nb_interior_points, replace=False)

    for i in initial_points:
        point = sample_simplex(i, simplex_size, 3)
        if np.any(np.isclose(point, 0., atol=1e-7)):
            continue
        sol = root(f, point, method="hybr")  # ,xtol=1.49012e-10,maxfev=1000
        if sol.success:
            v = sol.x / simplex_size
            if check_if_point_in_unit_simplex(v, delta):
                # only add new fixed points to list
                if not np.array([np.allclose(v, x, atol=atol) for x in roots]).any():
                    roots.append(v)

    return roots


def check_if_point_in_unit_simplex(point: np.ndarray, delta: Optional[float] = 1e-12) -> bool:
    """
    Checks if a point (in barycentric coordinates) is inside the unit simplex.

    Parameters
    ----------
    point: numpy.ndarray
        The barycentric coordinates of the point.
    delta: float
        Tolerance to consider a point outside the unit simplex.

    Returns
    -------
    bool
        Whether the point is inside the unit simplex.

    """
    if not np.isclose(np.sum(point), 1., atol=1.e-2):
        return False

    if not np.all((point > -delta) & (point < 1 + delta)):  # only if fp in simplex
        return False

    return True


def perturb_state(state: Union[Tuple[float, float, float], np.ndarray],
                  perturbation: Optional[float] = 0.01) -> List[np.ndarray]:
    """
    Produces a number of points in the simplex close to the state.

    If the sate is a vertex or in an edge, the perturbation is only made
    across the edges (we don't look for points in the interior of the simplex).

    Parameters
    ----------
    state: Union[Tuple[float, float, float], numpy.ndarray]
        Barycentric coordinates of a point inside the simplex.
    perturbation: float
        The amount of perturbation to apply to the point.

    Returns
    -------
    List[numpy.ndarray]
        A list of points (in barycentric coordinates) which are close to the state in the simplex.
    """
    new_states = []
    point_location, indexes = find_where_point_is_in_simplex(state)
    # first we check where the point is

    if point_location == 0:
        # If the point is a vertex, then we only perturb across that axis
        tmp1 = state.copy()
        tmp1[indexes[0]] -= perturbation
        tmp1[(indexes[0] + 1) % 3] += perturbation
        new_states.append(tmp1)
        tmp2 = state.copy()
        tmp2[indexes[0]] -= perturbation
        tmp2[(indexes[0] + 2) % 3] += perturbation
        new_states.append(tmp2)
    elif point_location == 1:
        # If the point in an edge, we will produce 2 points
        if state[indexes[0]] >= perturbation:
            tmp1 = state.copy()
            tmp1[indexes[0]] -= perturbation
            tmp1[indexes[1]] += perturbation
            new_states.append(tmp1)
        if state[indexes[1]] >= perturbation:
            tmp2 = state.copy()
            tmp2[indexes[0]] += perturbation
            tmp2[indexes[1]] -= perturbation
            new_states.append(tmp2)
    else:
        # if the point is in the interior of the simplex, we will
        # produce 3 points, each in the direction of a vertex
        for i in range(len(state)):
            tmp = state.copy()

            if tmp[i] <= 1. - perturbation:
                tmp[i] += perturbation
                if tmp[(i + 1) % 3] >= perturbation:
                    tmp[(i + 1) % 3] -= perturbation
                else:
                    tmp[(i + 2) % 3] -= perturbation
                new_states.append(tmp)

    return new_states


def perturb_state_discrete(state: Union[Tuple[float, float, float], np.ndarray],
                           size: int,
                           perturbation: Optional[int] = 1) -> List[np.ndarray]:
    """
    Produces a number of points in the simplex close to the state.

    If the sate is a vertex or in an edge, the perturbation is only made
    across the edges (we don't look for points in the interior of the simplex).

    Parameters
    ----------
    state: Union[Tuple[float, float, float], numpy.ndarray]
        The barycentric coordinates of a point inside the simplex.
    size: int
        The size of the edges of the simplex. This should coincide with the size of the finite population
        in Moran dynamics.
    perturbation: int
        The amount of perturbation to apply to the point.

    Returns
    -------
    List[numpy.ndarray]
        A list of points (in barycentric coordinates) which are close to the state in the simplex.
    """
    new_states = []
    point_location, indexes = find_where_point_is_in_simplex(state)
    # first we check where the point is

    if point_location == 0:
        # If the point is a vertex, then we only perturb across that axis
        tmp1 = state.copy()
        tmp1[indexes[0]] -= perturbation
        tmp1[(indexes[0] + 1) % 3] += perturbation
        new_states.append(tmp1)
        tmp2 = state.copy()
        tmp2[indexes[0]] -= perturbation
        tmp2[(indexes[0] + 2) % 3] += perturbation
        new_states.append(tmp2)
    elif point_location == 1:
        # If the point in an edge, we will produce 2 points
        if state[indexes[0]] >= perturbation:
            tmp1 = state.copy()
            if tmp1.sum() < size:
                tmp1[indexes[0]] += size - tmp1.sum()
            tmp1[indexes[0]] -= perturbation
            tmp1[indexes[1]] += perturbation
            new_states.append(tmp1)
        if state[indexes[1]] >= perturbation:
            tmp2 = state.copy()
            if tmp2.sum() < size:
                tmp2[indexes[0]] += size - tmp2.sum()
            tmp2[indexes[0]] += perturbation
            tmp2[indexes[1]] -= perturbation
            new_states.append(tmp2)
    else:
        # if the point is in the interior of the simplex, we will
        # produce 3 points, each in the direction of a vertex
        for i in range(len(state)):
            tmp = state.copy()
            if tmp.sum() < size:
                tmp[0] += size - tmp.sum()

            if tmp[i] + perturbation <= size:
                tmp[i] += perturbation
                if tmp[(i + 1) % 3] >= perturbation:
                    tmp[(i + 1) % 3] -= perturbation
                else:
                    tmp[(i + 2) % 3] -= perturbation
                new_states.append(tmp)

    return new_states


def add_arrow(line, position: Optional[float] = None, direction: Optional[str] = 'right', size: Optional[int] = 15,
              color: Optional[Union[str, Tuple[int, int, int]]] = None, arrowstyle: Optional[str] = '-|>',
              zorder: Optional[int] = 0):
    """
    add an arrow to a line.

    line:       Line2D object
    position:   x-position of the arrow. If None, mean of xdata is taken
    direction:  'left' or 'right'
    size:       size of the arrow in fontsize points
    color:      if None, line color is taken.
    """
    if color is None:
        color = line.get_color()

    xdata = line.get_xdata()
    ydata = line.get_ydata()

    if position is None:
        position = xdata.mean()
    # find closest index
    start_ind = np.argmin(np.absolute(xdata - position))
    if direction == 'right':
        end_ind = start_ind + 1
    else:
        end_ind = start_ind - 1

    line.axes.annotate('',
                       xytext=(xdata[start_ind], ydata[start_ind]),
                       xy=(xdata[end_ind], ydata[end_ind]),
                       arrowprops=dict(arrowstyle=arrowstyle, color=color),
                       size=size, zorder=zorder
                       )


def find_where_point_is_in_simplex(point: np.ndarray) -> Tuple[int, np.ndarray]:
    """
    Finds in which part of the 2D simplex the point is.

    This function will return:
    0 -> if the point is a vertex
    1 -> if the point in an edge
    2 -> if the point is in the interior of the simplex

    Parameters
    ----------
    point : numpy.ndarray
        The barycentric coordinates of the point

    Returns
    -------
    Tuple[int, numpy.ndarray]
        An integer which indicates where the point is in the simplex and
        the index of the non-zero entries.
    """
    # first we check if the root is in one of the edges
    if np.isclose(point, 0., atol=1e-7).any():
        # Then we check if it might be a vertex
        edge = np.where(~np.isclose(point, 0., atol=1e-7))[0]
        if edge.shape[0] > 1:  # we are at an edge
            return 1, edge
        else:  # we are at a vertex
            return 0, edge
    else:  # we are in the interior of the simplex
        return 2, np.array([0, 1, 2])



"""
The :mod:`egttools` package implements methods to study evolutionary dynamics.
"""
# Workaround for mac OSX
# import numpy

try:
    import egttools.numerical as numerical
except Exception:
    raise Exception("numerical package not initialized")
else:
    # Now we delete the pre-loaded numpy, as it is not necessary
    # del numpy
    from .numerical.numerical_ import __version__
    from .numerical.numerical_ import VERSION
    from .numerical.numerical_ import Random
    from .numerical.numerical_ import USES_BOOST
    from .numerical.numerical_ import (sample_simplex, sample_unit_simplex, calculate_nb_states,
                                       calculate_state,
                                       calculate_strategies_distribution, )

    import egttools.games as games
    import egttools.behaviors as behaviors
    import egttools.analytical as analytical
    import egttools.utils as utils
    import egttools.plotting as plotting
    import egttools.distributions as distributions
    import egttools.datastructures as datastructures

__all__ = ['utils', 'plotting', 'analytical',
           'games', 'behaviors', 'numerical',
           'distributions', 'datastructures', '__version__', 'VERSION', 'Random',
           'sample_simplex', 'sample_unit_simplex', 'calculate_nb_states', 'calculate_state',
           'calculate_strategies_distribution']



from abc import ABC, abstractmethod


class AbstractCPRStrategy(ABC):
    @abstractmethod
    def get_extraction(self, a: float, b: float, group_size: int, commitment: bool = False) -> float:
        pass

    @staticmethod
    @abstractmethod
    def get_payoff(a: float, b: float, extraction: float, group_extraction: float, fine: float = 0,
                   cost: float = 0, commitment: bool = False) -> float:
        pass

    @abstractmethod
    def would_like_to_commit(self) -> bool:
        pass

    @abstractmethod
    def proposes_commitment(self) -> bool:
        pass

    @abstractmethod
    def is_commitment_validated(self, nb_committers: int) -> bool:
        pass

    @abstractmethod
    def type(self) -> str:
        pass

    def __str__(self) -> str:
        return "AbstractCPRStrategy"



import numpy as np
import egttools as egt


def test_analytical_fitness_calculation_stoch_dynamics():
    group_size = 6
    min_nb_cooperators = 3
    c = 0.1
    b = 1
    population_size = 50
    risk = 1.
    game = egt.games.OneShotCRD(b, c, risk, group_size, min_nb_cooperators)
    payoffs = egt.utils.transform_payoffs_to_pairwise(game.nb_strategies(), game)
    evolver = egt.analytical.StochDynamics(game.nb_strategies(), payoffs, population_size, group_size)
    result = evolver.fitness_group(population_size - 1, 0, 1)

    assert result > 0


def test_analytical_pairwise_comparison_generic_game():
    group_size = 6
    min_nb_cooperators = 3
    c = 0.1
    b = 1
    population_size = 50
    risk = 1.
    game = egt.games.OneShotCRD(b, c, risk, group_size, min_nb_cooperators)
    evolver = egt.analytical.PairwiseComparison(population_size, game)
    result = evolver.calculate_fixation_probability(0, 1, 1)

    assert np.isclose(0.003682100470889826, result)


def test_analytical_pairwise_comparison_2_player_game():
    # Payoff matrix
    v = 2
    d = 3
    t = 1
    payoff_matrix = np.array([
        [(v - d) / 2, v],
        [0, (v / 2) - t],
    ])
    game = egt.games.Matrix2PlayerGameHolder(2, payoff_matrix=payoff_matrix)
    evolver = egt.analytical.PairwiseComparison(100, game)
    result = evolver.calculate_fixation_probability(0, 1, 1)

    assert np.isclose(0.8641155742462664, result)


def test_analytical_pairwise_comparison_n_player_game():
    # Payoff matrix
    v = 2
    d = 3
    t = 1
    payoff_matrix = np.array([
        [(v - d) / 2, v, v / 2, d],
        [0, (v / 2) - t, t, v],
    ])
    game = egt.games.MatrixNPlayerGameHolder(nb_strategies=2, group_size=3, payoff_matrix=payoff_matrix)
    evolver = egt.analytical.PairwiseComparison(100, game)
    result = evolver.calculate_fixation_probability(1, 0, 1)

    assert np.isclose(0.020104159364630978, result)


def test_if_stoch_dynamics_matches_with_pairwise_comparison():
    # Payoff matrix
    v = 2
    d = 3
    t = 1
    payoff_matrix = np.array([
        [(v - d) / 2, v],
        [0, (v / 2) - t],
    ])
    game = egt.games.Matrix2PlayerGameHolder(2, payoff_matrix=payoff_matrix)
    evolver1 = egt.analytical.PairwiseComparison(100, game)
    evolver2 = egt.analytical.StochDynamics(nb_strategies=2, payoffs=payoff_matrix, pop_size=100, group_size=2, mu=0)

    assert np.isclose(evolver1.calculate_fixation_probability(0, 1, 1), evolver2.fixation_probability(0, 1, 1))
    assert np.isclose(evolver1.calculate_gradient_of_selection(1, np.array([30
                                                                               , 70]))[0],
                      evolver2.gradient_selection(30, 0, 1, 1))
    transition1, fixation1 = evolver1.calculate_transition_and_fixation_matrix_sml(1)
    transition2, fixation2 = evolver2.transition_and_fixation_matrix(1)

    assert np.allclose(transition1.transpose(), transition2)
    assert np.allclose(fixation1, fixation2)

    full_transitions1 = evolver1.calculate_transition_matrix(1, 0.01)
    evolver2.mu = 0.01
    full_transitions2 = evolver2.calculate_full_transition_matrix(1)

    assert np.allclose(full_transitions1.toarray().transpose(), full_transitions2.toarray())

    # Now the same but for N-player games
    payoff_matrix = np.array([
        [(v - d) / 2, v, v / 2, d],
        [0, (v / 2) - t, t, v],
    ])
    game = egt.games.MatrixNPlayerGameHolder(nb_strategies=2, group_size=3, payoff_matrix=payoff_matrix)
    pairwise_payoffs = egt.utils.transform_payoffs_to_pairwise(2, game)
    evolver1 = egt.analytical.PairwiseComparison(100, game)
    evolver2 = egt.analytical.StochDynamics(nb_strategies=2, payoffs=pairwise_payoffs, pop_size=100, group_size=3, mu=0)

    assert np.isclose(evolver1.calculate_fixation_probability(1, 0, 1), evolver2.fixation_probability(1, 0, 1))
    assert np.isclose(evolver1.calculate_gradient_of_selection(1, np.array([30, 70]))[0],
                      evolver2.gradient_selection(30, 0, 1, 1))

    transition1, fixation1 = evolver1.calculate_transition_and_fixation_matrix_sml(1)
    transition2, fixation2 = evolver2.transition_and_fixation_matrix(1)

    assert np.allclose(transition1.transpose(), transition2)
    assert np.allclose(fixation1, fixation2)

    full_transitions1 = evolver1.calculate_transition_matrix(1, 0.01)
    evolver2 = egt.analytical.StochDynamics(nb_strategies=2, payoffs=payoff_matrix, pop_size=100, group_size=3, mu=0.01)
    full_transitions2 = evolver2.calculate_full_transition_matrix(1)

    assert np.allclose(full_transitions1.toarray().transpose(), full_transitions2.toarray())


def test_replicator_dynamics_2_player():
    # Payoff matrix
    v = 2
    d = 3
    t = 1
    payoff_matrix = np.array([
        [(v - d) / 2, v],
        [0, (v / 2) - t],
    ])
    x = np.array([0, 1])
    result = egt.analytical.replicator_equation(x, payoff_matrix)

    # Now the same but for N-player games
    payoff_matrix2 = np.array([
        [(v - d) / 2, v, v / 2, d],
        [0, (v / 2) - t, t, v],
    ])

    result = egt.analytical.replicator_equation_n_player(x, payoff_matrix2, group_size=3)




# Copyright (c) 2019-2022  Elias Fernandez
#
# This file is part of EGTtools.
#
# EGTtools is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# EGTtools is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with EGTtools.  If not, see <http://www.gnu.org/licenses/>

import matplotlib.colors
import matplotlib.pyplot as plt
import matplotlib.tri as tri
import numpy as np

from scipy.integrate import odeint
from matplotlib.patches import Circle
from typing import Optional, Tuple, List, Union, Callable, TypeVar
from numpy.typing import ArrayLike
from .. import (sample_simplex, sample_unit_simplex, calculate_nb_states, )
from .helpers import (barycentric_to_xy_coordinates, perturb_state, add_arrow,
                      perturb_state_discrete, find_where_point_is_in_simplex,
                      xy_to_barycentric_coordinates)

SelfSimplex2D = TypeVar("SelfSimplex2D", bound="Simplex2D")


class Simplex2D:
    top_corner = np.sqrt(3) / 2
    side_slope = np.sqrt(3)
    corners = np.array([
        [0, 0],
        [1 / 2, top_corner],
        [1, 0]
    ])
    triangle = tri.Triangulation(corners[:, 0], corners[:, 1])
    refiner = tri.UniformTriRefiner(triangle)
    trimesh = refiner.refine_triangulation(subdiv=5)

    def __init__(self, nb_points: Optional[int] = 1000, discrete: Optional[bool] = False,
                 size: Optional[Union[int, None]] = None):
        """
        Plots a 2-dimensional simplex in a cartesian plane.

        This class offers utility methods to plot gradients and equilibrium points on a 2-simplex (triangle).

        The plotting is always done on the unit simplex for convenience. At the moment no rotations are
        implemented, but we plan to add this feature, so that the triangle can be rotated before the plot.

        We discern between continuous and discrete dynamics. The main reason is that this class' objective
        is to plot evolutionary dynamics on a simplex. When we are working with the replicator equation
        it is straightforward to calculate all the gradients on the unit simplex. However, when working
        with finite populations using the social learning model (social imitation), we are actually working
        with a simplex with size equivalent to the population size (so all the dimensions of the simplex must
        sum to `Z`) and we only consider discrete (integer) values inside the simplex (the population may
        only have integer individuals). Of course this can be translated into frequencies, which gets us
        back to the unit simplex, but it is not so simple to transform any value between 0-1 sampled with
        numpy.linspace to a discrete value.

        Therefore, for the discrete case, will will sample directly discrete points in barycentric
        coordinates and only then, translate them into cartesian cooordinates.

        Parameters
        ----------
        nb_points : int
            number of points for which to calculate the gradients
        discrete : bool
            indicates whether we are in the continuous or discrete case
        size : int
            if we are in the discrete case, indicates the size of the simplex

        See Also
        --------
        egttools.plotting.plot_gradient,
        egttools.plotting.draw_invasion_diagram,
        egttools.analytical.replicator_equation,
        egttools.analytical.StochDynamics

        Cite
        -----
        This class has been inspired from: https://github.com/marvinboe/egtsimplex

        Examples
        --------
        >>> import numpy as np
        >>> import matplotlib.pyplot as plt
        >>> from egttools.plotting.helpers import (xy_to_barycentric_coordinates, calculate_stationary_points,
            ... calculate_stability)
        >>> from egttools.helpers.vectorized import (vectorized_replicator_equation,
            ... vectorized_barycentric_to_xy_coordinates)
        >>> from egttools.analytical import replicator_equation
        >>> simplex = Simplex2D()
        >>> payoffs = np.array([[1, 0, 0],
            ...        [0, 2, 0],
            ...        [0, 0, 3]])
        >>> v = np.asarray(xy_to_barycentric_coordinates(simplex.X, simplex.Y, simplex.corners))
        >>> results = vectorized_replicator_equation(v, payoffs)
        >>> xy_results = vectorized_barycentric_to_xy_coordinates(results, simplex.corners)
        >>> Ux = xy_results[:, :, 0].astype(np.float64)
        >>> Uy = xy_results[:, :, 1].astype(np.float64)
        >>> calculate_gradients = lambda u: replicator_equation(u, payoffs)
        >>> roots, roots_xy = calculate_stationary_points(simplex.trimesh.x, simplex.trimesh.y,
            ... simplex.corners, calculate_gradients)
        >>> stability = calculate_stability(roots, calculate_gradients)
        >>> type_labels = ['A', 'B', 'C']
        >>> fig, ax = plt.subplots(figsize=(10,8))
        >>> plot = (simplex.add_axis(ax=ax)
            ...            .apply_simplex_boundaries_to_gradients(Ux, Uy)
            ...            .draw_triangle()
            ...            .draw_gradients(zorder=0)
            ...            .add_colorbar()
            ...            .draw_stationary_points(roots_xy, stability)
            ...            .add_vertex_labels(type_labels)
            ...            .draw_trajectory_from_roots(lambda u, t: replicator_equation(u, payoffs),
            ...                                        roots,
            ...                                        stability,
            ...                                        trajectory_length=15,
            ...                                        linewidth=1,
            ...                                        step=0.01,
            ...                                        color='k', draw_arrow=True, arrowdirection='right',
            ...                                        arrowsize=30, zorder=4, arrowstyle='fancy')
            ...            .draw_scatter_shadow(lambda u, t: replicator_equation(u, payoffs), 300, color='gray',
            ...                                 marker='.', s=0.1, zorder=0)

        .. image:: ../images/simplex_example_infinite_pop_1.png

        >>> plot = (simplex.add_axis(ax=ax)
            ...            .apply_simplex_boundaries_to_gradients(Ux, Uy)
            ...            .draw_triangle()
            ...            .draw_stationary_points(roots_xy, stability)
            ...            .add_vertex_labels(type_labels)
            ...            .draw_trajectory_from_roots(lambda u, t: replicator_equation(u, payoffs),
            ...                                        roots,
            ...                                        stability,
            ...                                        trajectory_length=15,
            ...                                        linewidth=1,
            ...                                        step=0.01,
            ...                                        color='k', draw_arrow=True, arrowdirection='right',
            ...                                        arrowsize=30, zorder=4, arrowstyle='fancy')
            ...            .draw_scatter_shadow(lambda u, t: replicator_equation(u, payoffs), 300, color='gray',
            ...                                 marker='.', s=0.1, zorder=0)

        .. image:: ../images/simplex_example_infinite_pop_2.png
        """
        self.nb_points = nb_points
        self.Ux = None
        self.Uy = None
        self.stream = None
        x = np.linspace(0, 1, nb_points)
        y = np.linspace(0, 1, nb_points)
        self.X, self.Y = np.meshgrid(x, y)
        self.figure = None
        self.ax = None
        self.discrete = discrete
        self.size = size
        self.random_drift_edges = []

        if discrete:
            self.nb_states = calculate_nb_states(self.size, 3)
            xy_coords = []

            for i in range(self.nb_states):
                state = sample_simplex(i, size, 3)
                xy_coords.append(barycentric_to_xy_coordinates(state / size, self.corners))
            xy_coords = np.asarray(xy_coords)

            self.triangle_discrete = tri.Triangulation(xy_coords[:, 0].flatten(), xy_coords[:, 1].flatten())

    def add_axis(self, figsize: Optional[Tuple[int, int]] = (10, 8), ax: Optional[plt.axis] = None) -> SelfSimplex2D:
        """
        Creates or stores a new axis inside the class.

        Parameters
        ----------
        figsize: Optional[Tuple[int, int]]
            The size of the figure. This argument is only used if no ax is given.
        ax: Optional[matplotlib.pyplot.axis]
            If given, the axis will be stored inside the object. Otherwise, a new axis will be created.

        Returns
        -------
        Simplex2D
            The class object.
        """
        if ax is not None:
            self.ax = ax
            self.figure = ax.figure
        else:
            self.figure, self.ax = plt.subplots(figsize=figsize)

        return self

    def add_edges_with_random_drift(self, random_drift_edges: List[Tuple[int, int]]) -> SelfSimplex2D:
        """
        Adds information to the class about which edges have random drift.

        This will be used to avoid plotting a lot equilibria alongside an edge.

        Parameters
        ----------
        random_drift_edges: List[Tuple[int, int]]
            A list of tuples which indicate the (undirected) edges in which there is random drift.

        Returns
        -------
        Simplex2D
            The class object.
        """
        self.random_drift_edges = random_drift_edges
        return self

    def get_figure_and_axis(self) -> Tuple[plt.figure, plt.axis]:
        """
        Returns the stored figure and axis.

        Returns
        -------
        Tuple[matplotlib.pyplot.figure, matplotlib.pyplot.axis]
            The figure and axis stored in the current object.
        """
        return self.figure, self.ax

    def apply_simplex_boundaries_to_gradients(self, u: np.ndarray, v: np.ndarray) -> SelfSimplex2D:
        """
        Applies boundaries of the triangle to a list of gradient values over the cartesian grid.

        The boundaries are applied using the X Y grid defined in the instantiation of the class.

        Parameters
        ----------
        u: numpy.ndarray
            The X component of the gradients.
        v: numpy.ndarray
            The Y component of the gradients
        Returns
        -------
        Simplex2D
            A reference to the class object.
        """
        self.Ux = u.copy()
        self.Uy = v.copy()

        self.Ux = np.where(self.X >= self.Y / self.side_slope, self.Ux, np.nan)
        self.Ux = np.where(self.Y <= -self.side_slope * self.X + self.side_slope, self.Ux, np.nan)
        self.Ux = np.where(self.Y <= self.top_corner, self.Ux, np.nan)
        self.Ux = np.where(self.Y > 0, self.Ux, np.nan)
        self.Uy = np.where(self.X >= self.Y / self.side_slope, self.Uy, np.nan)
        self.Uy = np.where(self.Y <= -self.side_slope * self.X + self.side_slope, self.Uy, np.nan)
        self.Uy = np.where(self.Y <= self.top_corner, self.Uy, np.nan)
        self.Uy = np.where(self.Y > 0, self.Uy, np.nan)

        return self

    def draw_triangle(self, color: Optional[str] = 'k', linewidth: Optional[int] = 2,
                      linewidth_random_drift: Optional[int] = 4) -> SelfSimplex2D:
        """
        Draws the borders of a triangle enclosing the 2-simplex.

        Parameters
        ----------
        color: Optional[str]
            The color of the borders of the triangle.
        linewidth: Optional[int]
            The width of the borders of the triangle.
        linewidth_random_drift: Optional[int]
            The width of the dashed line that represents the edges with random drift.

        Returns
        -------
        Simplex2D
            A refernece to the class object.
        """
        self.ax.triplot(self.triangle, color=color, linewidth=linewidth)
        for edge in self.random_drift_edges:
            self.ax.plot([self.corners[edge[0], 0], self.corners[edge[1], 0]],
                         [self.corners[edge[0], 1], self.corners[edge[1], 1]], lw=linewidth_random_drift,
                         linestyle='dashed', color=color)
        return self

    def draw_gradients(self, arrowsize: Optional[float] = 2,
                       arrowstyle: Optional[str] = 'fancy',
                       color: Optional[Union[str, Tuple[int, int, int]]] = None, density: Optional[float] = 1,
                       linewidth: Optional[float] = 1.5,
                       cmap: Optional[Union[str, matplotlib.colors.Colormap]] = 'viridis',
                       zorder: Optional[int] = 0) -> SelfSimplex2D:
        """
        Draws the gradients inside the unit simplex using a streamplot.

        Parameters
        ----------
        arrowsize: Optional[float]
            The size of the arrows of the gradients
        arrowstyle: Optional[str]
            The style of the arrows. See matplotlib arrowstyles.
        color: Optional[Union[str, Tuple[int, int, int]]]
            The color of the arrows. If no color is given, it will be generated as a function of the gradients.
        density: Optional[float]
            The density of arrows (how many arrows) to plot.
        linewidth: Optional[float]
            The width of the arrows.
        cmap: Optional[Union[str, matplotlib.colors.Colormap]]
            The color map to be used.
        zorder: Optional[int]
            The order in which the gradients should appear in the plot (above or below other elements).

        Returns
        -------
        Simplex2D
            A reference to the class object.

        """
        if self.Ux is None or self.Uy is None:
            raise Exception("Please call Simplex.apply_simplex_boundaries_to_gradients first")

        velocities = (self.Ux ** 2 + self.Uy ** 2) ** 0.5
        colors = velocities if color is None else color

        self.stream = self.ax.streamplot(self.X, self.Y,
                                         self.Ux, self.Uy,
                                         arrowsize=arrowsize,
                                         arrowstyle=arrowstyle,
                                         color=colors,
                                         density=density,
                                         linewidth=linewidth,
                                         cmap=cmap,
                                         zorder=zorder
                                         )
        return self

    def add_colorbar(self, aspect: Optional[float] = 10,
                     anchor: Optional[Tuple[float, float]] = (-0.5, 0.5),
                     panchor: Optional[Tuple[float, float]] = (0, 0),
                     shrink: Optional[float] = 0.6,
                     label: Optional[str] = 'gradient of selection',
                     label_rotation: Optional[int] = 270,
                     label_fontsize: Optional[int] = 16,
                     labelpad: Optional[float] = 20) -> SelfSimplex2D:
        """
        Adds a color bar to indicate the meaning of the colors of the plotted gradients.
        This should only be used if the gradients were plotted and the colors have been drawn in function
        of the strength of the gradient.

        Parameters
        ----------
        aspect: Optional[float]
            Aspect ration of the color bar.
        anchor: Optional[Tuple[float, float]]
            Anchor point for the color bar.
        panchor: Optional[Tuple[float, float]]
        shrink: Optional[float]
            Ration for shrinking the color bar.
        label: Optional[str]
            Label for the color bar.
        label_rotation: Optional[int]
            Rotation of the label.
        label_fontsize: Optional[int]
            Font size of the label.
        labelpad: Optional[float]
            How much padding should be added to the label.

        Returns
        -------
        Simplex2D
            A reference to the class object.

        """
        cbar = self.figure.colorbar(self.stream.lines, aspect=aspect, anchor=anchor, panchor=panchor, shrink=shrink,
                                    ax=self.ax)
        cbar.set_label(label, rotation=label_rotation, fontsize=label_fontsize, labelpad=labelpad)

        return self

    def draw_stationary_points(self, roots: List[Union[Tuple[float, float], np.ndarray]],
                               stability: Union[List[bool], List[int]],
                               zorder: Optional[int] = 5, linewidth: Optional[float] = 3,
                               atol: Optional[float] = 1e-7) -> SelfSimplex2D:
        """
        Draws the black circles for stable points and white circles for unstable ones.

        Parameters
        ----------
        roots:
            A list of arrays (or tuples) containing the cartesian coordinates of the roots.
        stability:
            A list of boolean or integer values indicating whether the root is stable. If there are integer values
            -1 - unstable, 0 - saddle, 1 - stable.
        zorder:
            Indicates in which order these points should appear in the figure (above or below other plots).
        linewidth:
            Width of the border of the circles that represents the roots.
        atol:
            Tolerance to consider a value equal to 0. Used to check if a point is on an edge.

        Returns
        -------
        Simplex2D
            A reference to the class object.

        """
        for i, stationary_point in enumerate(roots):
            point = xy_to_barycentric_coordinates(stationary_point[0], stationary_point[1], self.corners)
            place, _ = find_where_point_is_in_simplex(point)
            if place == 1:
                # First let's check if stationary point is in an edge with random drift
                if np.isclose([point[3 - np.sum(edge)] for edge in self.random_drift_edges], 0.,
                              atol=atol).any():
                    continue

            if stability[i] == 1:
                facecolor = 'k'
            elif stability[i] == -1:
                facecolor = 'white'
            else:
                facecolor = 'grey'
            self.ax.add_artist(Circle(stationary_point, 0.015,
                                      edgecolor='k', facecolor=facecolor, zorder=zorder, linewidth=linewidth))
        return self

    def add_vertex_labels(self, labels: Union[Tuple[str, str, str], List[str]], epsilon_bottom: Optional[float] = 0.05,
                          epsilon_top: Optional[float] = 0.05,
                          fontsize: Optional[float] = 16,
                          horizontalalignment: Optional[str] = 'center') -> SelfSimplex2D:
        """
        Adds labels to the vertices of the triangle that represents the 2-simplex.

        Parameters
        ----------
        labels: Union[Tuple[str, str, str], List[str]]
            A tuple or a list containing 3 strings that give name to the vertices of the triangle. The order is
            bottom left corner, top corner, bottom right corner.
        epsilon_bottom: Optional[float]
            How much separation should the label have from the bottom vertices
        epsilon_top: Optional[float]
            How much separation should the label have from the top vertex.
        fontsize: Optional[float]
            Font size for the labels.
        horizontalalignment: Optional[str]
            Horizontal alignment for the label text.

        Returns
        -------
        Simplex2D
            A reference to the current object.

        """
        self.ax.annotate(labels[0], self.corners[0], xytext=self.corners[0] + np.array([-epsilon_bottom, 0.0]),
                         horizontalalignment=horizontalalignment, va='center', fontsize=fontsize)
        self.ax.annotate(labels[1], self.corners[1], xytext=self.corners[1] + np.array([0.0, epsilon_top + 0.02]),
                         horizontalalignment=horizontalalignment, va='top', fontsize=fontsize)
        self.ax.annotate(labels[2], self.corners[2], xytext=self.corners[2] + np.array([epsilon_bottom, 0.0]),
                         horizontalalignment=horizontalalignment, va='center', fontsize=fontsize)

        return self

    def draw_trajectories(self, f: Callable[[np.ndarray, int], np.ndarray], nb_trajectories: int,
                          trajectory_length: Optional[int] = 15, step: Optional[float] = 0.01,
                          color: Optional[Union[str, Tuple[int, int, int]]] = 'whitesmoke',
                          ms: Optional[float] = 0.5, zorder: Optional[int] = 0) -> SelfSimplex2D:
        """
        Draws trajectories inside the unit simplex starting from random initial points.

        Parameters
        ----------
        f: Callable[[np.ndarray, int], np.ndarray]
            Function that can calculate the gradient at any point in the simplex.
        nb_trajectories: int
            Number of trajectories to draw.
        trajectory_length: Optional[int]
            Length of the trajectory. This is used to calculate the amount of points odeint should calculate.
        step: Optional[float]
            The step size in time to get to the maximum trajectory length. Together with trajectory_length
            this indicates the amount of points odeint should calculate.
        color: Optional[Union[str, Tuple[int, int, int]]]
            The color of the points of the trajectory.
        ms: Optional[float]
            The size of the points.
        zorder: Optional[int]
            The order in which this plot should appear in the figure (above or bellow other plots).

        Returns
        -------
        Simplex2D
            A reference to the current object.

        """
        if self.discrete:
            if nb_trajectories > self.nb_states:
                nb_trajectories = self.nb_states
            initial_points = np.random.choice(range(self.nb_states), size=nb_trajectories, replace=False)
            for point in initial_points:
                u = sample_simplex(point, self.size, 3)
                x = odeint(f, u, np.arange(0, trajectory_length, step), full_output=False)
                # noinspection PyTypeChecker
                v = barycentric_to_xy_coordinates(x / self.size, self.corners)
                self.ax.plot(v[:, 0], v[:, 1], color, ms=ms, zorder=zorder)
        else:
            for _ in range(nb_trajectories):
                u = sample_unit_simplex(3)
                x = odeint(f, u, np.arange(0, trajectory_length, step), full_output=False)
                # noinspection PyTypeChecker
                v = barycentric_to_xy_coordinates(x, self.corners)
                self.ax.plot(v[:, 0], v[:, 1], color, ms=ms, zorder=zorder)

        return self

    def draw_trajectory_from_points(self, f: Callable[[np.ndarray, int], np.ndarray], points: List[np.ndarray],
                                    trajectory_length: Optional[int] = 15, step: Optional[float] = 0.1,
                                    color: Optional[Union[str, Tuple[int, int, int]]] = 'k',
                                    linewidth: Optional[float] = 0.5, zorder: Optional[int] = 0,
                                    draw_arrow: Optional[bool] = False, arrowstyle: Optional[str] = 'fancy',
                                    arrowsize: Optional[int] = 50,
                                    position: Optional[int] = None,
                                    arrowdirection: Optional[str] = 'right') -> SelfSimplex2D:
        """
        Draws trajectories inside the unit simplex starting from the indicated points.

        Parameters
        ----------
        f: Callable[[np.ndarray, int], np.ndarray]
            Function that can calculate the gradient at any point in the simplex.
        points: List[np.ndarray[np.float64[3,m]]
            A list of points in barycentric coordinates from which the trajectories should start.
        trajectory_length: Optional[int]
            Length of the trajectory. This is used to calculate the amount of points odeint should calculate.
        step: Optional[float]
            The step size in time to get to the maximum trajectory length. Together with trajectory_length
            this indicates the amount of points odeint should calculate.
        color: Optional[Union[str, Tuple[int, int, int]]]
            The color of the points of the trajectory.
        linewidth: Optional[float] = 0.5
            Width of the line to be plot.
        zorder: Optional[int]
            The order in which this plot should appear in the figure (above or bellow other plots).
        draw_arrow: Optional[bool]
            Indicates whether to draw an arrow along the trajectory.
        arrowstyle: Optional[str]
            Indicates the style of the arrow to be plotted.
        arrowsize: Optional[int]
            The size of the arrow.
        position: Optional[int]
            Where should the arrow be pltoted.
        arrowdirection: Optional[str]
            Indicates whether the arrow should be plotted in the direction of the advancing trajectory (right) or
            the opposite.

        Returns
        -------
        Simplex2D
            A reference to the current object.
        """

        for i, point in enumerate(points):
            x = odeint(f, point, np.arange(0, trajectory_length, step), full_output=False)

            # noinspection PyTypeChecker
            v = barycentric_to_xy_coordinates(x, self.corners)
            line = self.ax.plot(v[:, 0], v[:, 1], color, linewidth=linewidth, zorder=zorder)[0]
            if draw_arrow:
                add_arrow(line, size=arrowsize, arrowstyle=arrowstyle, position=position,
                          direction=arrowdirection)

        return self

    def draw_trajectory_from_roots(self, f: Callable[[np.ndarray, int], np.ndarray], roots: List[np.ndarray],
                                   stability: List[bool],
                                   trajectory_length: Optional[int] = 15, step: Optional[float] = 0.1,
                                   perturbation: Optional[Union[int, float]] = 0.01,
                                   color: Optional[Union[str, Tuple[int, int, int]]] = 'k',
                                   linewidth: Optional[float] = 0.5, zorder: Optional[int] = 0,
                                   draw_arrow: Optional[bool] = False, arrowstyle: Optional[str] = 'fancy',
                                   arrowsize: Optional[int] = 50,
                                   position: Optional[int] = None,
                                   arrowdirection: Optional[str] = 'right',
                                   atol: Optional[float] = 1e-7) -> SelfSimplex2D:
        """
        Draws trajectories inside the unit simplex starting from the stationary points.

        Parameters
        ----------
        f: Callable[[np.ndarray, int], np.ndarray]
            Function that can calculate the gradient at any point in the simplex.
        roots: List[np.ndarray[np.float64[3,m]]
            A list of points in barycentric coordinates from which the trajectories should start.
        stability: List[bool]
            Indicates whether the root is a stable or unstable point.
        trajectory_length: Optional[int]
            Length of the trajectory. This is used to calculate the amount of points odeint should calculate.
        step: Optional[float]
            The step size in time to get to the maximum trajectory length. Together with trajectory_length
            this indicates the amount of points odeint should calculate.
        perturbation: Optional[Union[int, float]]
            Indicates how much perturbation should be applied to the root to start drawing the trajectory.
            If no perturbation is applied, since the gradient is 0, the system will never leave the root.
        color: Optional[Union[str, Tuple[int, int, int]]]
            The color of the points of the trajectory.
        linewidth: Optional[float] = 0.5
            Width of the line to be plot.
        zorder: Optional[int]
            The order in which this plot should appear in the figure (above or bellow other plots).
        draw_arrow: Optional[bool]
            Indicates whether to draw an arrow along the trajectory.
        arrowstyle: Optional[str]
            Indicates the style of the arrow to be plotted.
        arrowsize: Optional[int]
            The size of the arrow.
        position: Optional[int]
            Where should the arrow be pltoted.
        arrowdirection: Optional[str]
            Indicates whether the arrow should be plotted in the direction of the advancing trajectory (right) or
            the opposite.
        atol: Optional[float]
            Tolerance to consider a value equal to 0. Used to check if a point is on an edge of the simplex.

        Returns
        -------
        Simplex2D
            A reference to the current object.
        """
        if self.discrete:
            if type(perturbation) is float:
                perturbation = 1

            for i, stationary_point in enumerate(roots):
                # First let's check if stationary point is in an edge with random drift
                if np.isclose([stationary_point[3 - np.sum(edge)] for edge in self.random_drift_edges], 0.,
                              atol=atol).any():
                    continue
                if stability[i]:  # we don't plot arrows starting at stable points
                    continue
                stationary_point_discrete = (stationary_point * self.size)
                states = perturb_state_discrete(stationary_point_discrete, self.size, perturbation=perturbation)
                for state in states:
                    x = odeint(f, state, np.arange(0, trajectory_length, step), full_output=False)
                    # check if point is outside simplex
                    # noinspection PyUnresolvedReferences
                    tmp_check = np.where(~np.isclose(x.sum(axis=1) / self.size, 1., atol=1e-2))[0]
                    if len(tmp_check) > 0:
                        last_inside = np.min(tmp_check)
                        if last_inside == 0:
                            continue
                        x = x[:last_inside]
                    # noinspection PyTypeChecker
                    v = barycentric_to_xy_coordinates(x / self.size, self.corners)
                    line = self.ax.plot(v[:, 0], v[:, 1], color, linewidth=linewidth, zorder=zorder)[0]
                    if draw_arrow:
                        add_arrow(line, size=arrowsize, arrowstyle=arrowstyle, position=position,
                                  direction=arrowdirection)
        else:
            for i, stationary_point in enumerate(roots):
                # First let's check if stationary point is in an edge with random drift
                if np.isclose([stationary_point[3 - np.sum(edge)] for edge in self.random_drift_edges], 0.,
                              atol=atol).any():
                    continue
                if stability[i] == 1:  # we don't plot arrows starting at stable points
                    continue
                states = perturb_state(stationary_point, perturbation=perturbation)
                for state in states:
                    x = odeint(f, state, np.arange(0, trajectory_length, step), full_output=False)

                    # noinspection PyTypeChecker
                    v = barycentric_to_xy_coordinates(x, self.corners)
                    line = self.ax.plot(v[:, 0], v[:, 1], color, linewidth=linewidth, zorder=zorder)[0]
                    if draw_arrow:
                        add_arrow(line, size=arrowsize, arrowstyle=arrowstyle, position=position,
                                  direction=arrowdirection)

        return self

    def draw_trajectory_from_vector(self, trajectory: np.ndarray,
                                    color: Optional[Union[str, Tuple[int, int, int]]] = 'k',
                                    linewidth: Optional[float] = 0.5, zorder: Optional[int] = 0, ):
        points = np.asarray([barycentric_to_xy_coordinates(x, self.corners) for x in trajectory])

        self.ax.plot(points[:, 0], points[:, 1], color, linewidth=linewidth, zorder=zorder)

        return self

    def draw_scatter_shadow(self, f: Callable[[np.ndarray, int], np.ndarray], nb_trajectories: int,
                            trajectory_length: Optional[int] = 15, step: Optional[float] = 0.1,
                            s: Optional[Union[float, ArrayLike]] = 0.1,
                            color: Optional[Union[str, Tuple[int, int, int]]] = 'whitesmoke',
                            marker: Optional[str] = '.', zorder: Optional[int] = 0) -> SelfSimplex2D:
        """
        Draws a series of point which follows trajectories in the simplex starting from random points.

        The visual effect is as if there were shadows in the direction of the gradient.

        Parameters
        ----------
        f: Callable[[np.ndarray, int], np.ndarray]
            Function that can calculate the gradient at any point in the simplex.
        nb_trajectories: int
            Number of trajectories to draw.
        trajectory_length: Optional[int]
            Length of the trajectory. This is used to calculate the amount of points odeint should calculate.
        step: Optional[float]
            The step size in time to get to the maximum trajectory length. Together with trajectory_length
            this indicates the amount of points odeint should calculate.
        s: Optional[Union[str, Tuple[int, int, int]]]
            Size of the points.
        color: Optional[Union[str, Tuple[int, int, int]]]
            The color of the points of the trajectory.
        marker: Optional[str]
            Style of the points to be drawn. See matplotlib markers.
        zorder: Optional[int]
            The order in which this plot should appear in the figure (above or bellow other plots).

        Returns
        -------
        Simplex2D
            A reference to the current object.
        """

        if self.discrete:
            if nb_trajectories > self.nb_states:
                nb_trajectories = self.nb_states
            initial_points = np.random.choice(range(self.nb_states), size=nb_trajectories, replace=False)
            for point in initial_points:
                u = sample_simplex(point, self.size, 3)
                x = odeint(f, u, np.arange(0, trajectory_length, step), full_output=False)
                # noinspection PyTypeChecker
                v = barycentric_to_xy_coordinates(x / self.size, self.corners)
                self.ax.scatter(v[:, 0], v[:, 1], s, color=color, marker=marker, zorder=zorder)
        else:
            for _ in range(nb_trajectories):
                u = sample_unit_simplex(3)
                x = odeint(f, u, np.arange(0, trajectory_length, step), full_output=False)
                # noinspection PyTypeChecker
                v = barycentric_to_xy_coordinates(x, self.corners)
                self.ax.scatter(v[:, 0], v[:, 1], s, color=color, marker=marker, zorder=zorder)

        return self

    def draw_stationary_distribution(self, stationary_distribution: np.ndarray,
                                     cmap: Optional[Union[str, matplotlib.colors.Colormap]] = 'binary',
                                     shading: Optional[str] = 'gouraud',
                                     alpha: Optional[float] = 1., edgecolors: Optional[str] = 'grey',
                                     vmin: Optional[float] = None, vmax: Optional[float] = None,
                                     zorder: Optional[int] = 0,
                                     colorbar: Optional[bool] = True,
                                     aspect: Optional[float] = 10,
                                     anchor: Optional[Tuple[float, float]] = (-0.5, 0.5),
                                     panchor: Optional[Tuple[float, float]] = (0, 0),
                                     shrink: Optional[float] = 0.6,
                                     label: Optional[str] = 'stationary distribution',
                                     label_rotation: Optional[int] = 270,
                                     label_fontsize: Optional[int] = 16,
                                     labelpad: Optional[float] = 20):
        """
        Draws the stationary distribution inside the simplex using a matplotlib.pyplot.tripcolor

        Parameters
        ----------
        stationary_distribution: numpy.ndarray
            An array containing the values of the stationary distribution. The order of these points
            must follow the order given by egttools.sample_simplex when iterating from 0-nb_states.
        cmap: Optional[Union[str, matplotlib.colors.Colormap]]
            Color map to be used.
        shading: Optional[str]
            Type of shading to be used in the plot. Can be either "gouraud" or "flat".
        alpha: Optional[float]
            The level of transparency.
        edgecolors: Optional[str]
            The colors of the edges of the triangular grid.
        vmin: Optional[flaot]
            The minimum value to take into account for the color range to plot.
        vmax: Optional[float]
            The maximum value to take into account for the color range to plot.
        zorder: Optional[int]
            The order in which this plot should appear in the figure (above or bellow other plots).
        colorbar: Optional[bool] = True
            Indicates whether to add a color bar to the plot.
        aspect: Optional[float]
            The aspect ration of the color bar.
        anchor: Optional[Tuple[float, float]]
            The anchor of the color bar.
        panchor: Optional[Tuple[float, float]]
            The panchor of the colorbar
        shrink: Optional[float]
            Ratio of shrinking the color bar.
        label: Optional[str]
            Label of the color bar.
        label_rotation: Optional[int]
            Rotation of the label.
        label_fontsize: Optional[int]
            Font size of the label.
        labelpad: Optional[float]
            How much padding should be added to the label.


        Returns
        -------
        Simplex2D
            A reference to the current object
        """
        if not self.discrete:
            raise Exception("The stationary distribution only exists in Finite populations modeled as a Markov Chain.")

        if shading != 'flat' and shading != 'gouraud':
            raise Exception("Shading can only the be 'flat' or 'gouraud'")

        sd_plot = self.ax.tripcolor(self.triangle_discrete, stationary_distribution, cmap=cmap, shading=shading,
                                    alpha=alpha, edgecolor=edgecolors,
                                    vmin=vmin, vmax=vmax,
                                    zorder=zorder)

        if colorbar:
            cbar = self.figure.colorbar(sd_plot, aspect=aspect, anchor=anchor, panchor=panchor, shrink=shrink,
                                        ax=self.ax)
            cbar.set_label(label, rotation=label_rotation, fontsize=label_fontsize, labelpad=labelpad)

        return self



from pytest import fixture
import numpy as np
from egttools.analytical import StochDynamics
from egttools import calculate_nb_states


@fixture
def test_analytical_fitness_calculation_stoch_dynamics():
    payoff_matrix = np.array([
        [-0.5, 2],
        [0, 0]
    ])

    evolver = StochDynamics(nb_strategies=2, payoffs=payoff_matrix, pop_size=50, group_size=2, mu=0)

    return evolver


def test_update_population_size(test_analytical_fitness_calculation_stoch_dynamics):
    evolver = test_analytical_fitness_calculation_stoch_dynamics

    evolver.update_population_size(100)

    assert evolver.pop_size == 100
    assert evolver.nb_states_population == calculate_nb_states(100, 2)


def test_update_group_size(test_analytical_fitness_calculation_stoch_dynamics):
    evolver = test_analytical_fitness_calculation_stoch_dynamics

    evolver.update_group_size(4)

    assert evolver.group_size == 4
    assert evolver.nb_group_combinations == calculate_nb_states(4, 2)
    assert evolver.fitness == evolver.fitness_group
    assert evolver.full_fitness == evolver.full_fitness_difference_group


def test_update_payoffs(test_analytical_fitness_calculation_stoch_dynamics):
    evolver = test_analytical_fitness_calculation_stoch_dynamics

    payoff_matrix = np.array([
        [-0.5, 2],
        [0, 5]
    ])

    evolver.update_payoffs(payoff_matrix)

    assert np.allclose(payoff_matrix, evolver.payoffs)

    payoff_matrix = np.array([
        [-0.5, 2, 3],
        [0, 5, 3],
        [0, 5, 3]
    ])

    try:
        evolver.update_payoffs(payoff_matrix)
    except ValueError:
        assert True
    else:
        assert False

    evolver.update_payoffs(payoff_matrix, 3)

    assert np.allclose(payoff_matrix, evolver.payoffs)


@fixture
def get_fitness(test_analytical_fitness_calculation_stoch_dynamics):
    evolver = test_analytical_fitness_calculation_stoch_dynamics

    fitness_i = ((evolver.pop_size - 1) / (evolver.pop_size - 1)) * evolver.payoffs[0, 1]

    fitness_j = ((1 / (evolver.pop_size - 1)) * evolver.payoffs[1, 1]) + (
            ((evolver.pop_size - 2) / (evolver.pop_size - 1)) * evolver.payoffs[1, 0])

    return fitness_i - fitness_j


def test_fitness_pair(test_analytical_fitness_calculation_stoch_dynamics, get_fitness):
    evolver = test_analytical_fitness_calculation_stoch_dynamics
    fitness_diff = get_fitness

    fitness = evolver.fitness_pair(1, 0, 1)

    assert np.isclose(fitness, fitness_diff)


def test_prob_increase_decrease(test_analytical_fitness_calculation_stoch_dynamics, get_fitness):
    evolver = test_analytical_fitness_calculation_stoch_dynamics
    fitness_diff = get_fitness

    prob_increase, prob_decrease = evolver.prob_increase_decrease(1, 0, 1, 1)

    increase = (((evolver.pop_size - 1) / evolver.pop_size) *
                (1 / (evolver.pop_size - 1))) * StochDynamics.fermi(-1, fitness_diff)

    decrease = ((1 / evolver.pop_size) *
                ((evolver.pop_size - 1) / (evolver.pop_size - 1))) * StochDynamics.fermi(1, fitness_diff)

    assert np.isclose(increase, prob_increase)
    assert np.isclose(decrease, prob_decrease)

    prob_increase, prob_decrease = evolver.prob_increase_decrease(0, 0, 1, 1)

    assert np.isclose(prob_increase, 0)
    assert np.isclose(prob_decrease, 0)

    prob_increase, prob_decrease = evolver.prob_increase_decrease(evolver.pop_size, 0, 1, 1)

    assert np.isclose(prob_increase, 0)
    assert np.isclose(prob_decrease, 0)


def test_prob_increase_decrease_with_mutation(test_analytical_fitness_calculation_stoch_dynamics, get_fitness):
    evolver = test_analytical_fitness_calculation_stoch_dynamics
    fitness_diff = get_fitness

    mu = 0.01
    evolver.mu = mu
    prob_increase, prob_decrease = evolver.prob_increase_decrease_with_mutation(1, 0, 1, 1)

    increase = (((evolver.pop_size - 1) / evolver.pop_size) *
                (1 / (evolver.pop_size - 1))) * StochDynamics.fermi(-1, fitness_diff)
    increase = ((1 - mu) * increase) + (mu * ((evolver.pop_size - 1) / evolver.pop_size))

    decrease = ((1 / evolver.pop_size) *
                ((evolver.pop_size - 1) / (evolver.pop_size - 1))) * StochDynamics.fermi(1, fitness_diff)
    decrease = ((1 - mu) * decrease) + (mu * (1 / evolver.pop_size))

    assert np.isclose(increase, prob_increase)
    assert np.isclose(decrease, prob_decrease)

    prob_increase, prob_decrease = evolver.prob_increase_decrease_with_mutation(0, 0, 1, 1)

    assert np.isclose(prob_increase, mu)
    assert np.isclose(prob_decrease, 0)

    prob_increase, prob_decrease = evolver.prob_increase_decrease_with_mutation(evolver.pop_size, 0, 1, 1)

    assert np.isclose(prob_increase, 0)
    assert np.isclose(prob_decrease, mu)



import sys

from widget_module import Widget


class DerivedWidget(Widget):
    def __init__(self, message):
        super().__init__(message)

    def the_answer(self):
        return 42

    def argv0(self):
        return sys.argv[0]



# -*- coding: utf-8 -*-
#
# Configuration file for the Sphinx documentation builder.
#
# This file does only contain a selection of the most common options. For a
# full list see the documentation:
# http://www.sphinx-doc.org/en/master/config

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
# import os
# import sys
# sys.path.insert(0, os.path.abspath('.'))

import git
import github
import requests

import io
import os
import subprocess
import sys
import tempfile
import zipfile

on_rtd = os.environ.get('READTHEDOCS') == 'True'

# -- General configuration ---------------------------------------------------

# If your documentation needs a minimal Sphinx version, state it here.
#
# needs_sphinx = '1.0'

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
sys.path.insert(1, os.path.abspath(os.path.dirname(__file__)))
extensions = [
    'sphinx.ext.napoleon',
    'sphinx.ext.autodoc',
    'sphinx.ext.autosummary',
    'sphinx.ext.doctest',
    'sphinx.ext.intersphinx',
    'sphinx.ext.todo',
    'sphinx.ext.coverage',
    'sphinx.ext.mathjax',
    'sphinx.ext.ifconfig',
    'sphinx.ext.viewcode',
    'sphinx.ext.githubpages',
    'sphinx_autodoc_typehints',  # Automatically document param types (less noise in class signature)
    'recommonmark',
    'nbsphinx',
    'pybind11_docstrings',
    'sphinxcontrib.bibtex',
]

bibtex_bibfiles = ['references.bib']
bibtex_encoding = 'latin'
bibtex_default_style = 'unsrt'

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# The suffix(es) of source filenames.
# You can specify multiple suffix as a list of string:
#
# source_suffix = ['.rst', '.md']
source_suffix = '.rst'

# The master toctree document.
master_doc = 'index'

# -- Project information -----------------------------------------------------

project = 'EGTtools'
copyright = '2019-2022, Elias Fernández'
author = 'Elias Fernández'

if on_rtd:
    rtd_version = os.environ.get('READTHEDOCS_VERSION')
    branch = 'docs' if rtd_version == 'latest' else rtd_version

    github_token = os.environ['GITHUB_TOKEN']
    head_sha = git.Repo(search_parent_directories=True).head.commit.hexsha
    g = github.Github()
    runs = g.get_repo('Socrats/EGTTools').get_workflow("wheels.yml").get_runs(branch=branch)
    artifacts_url = next(r for r in runs if r.head_sha == head_sha).artifacts_url

    archive_download_url = \
        next(artifact for artifact in requests.get(artifacts_url).json()['artifacts'] if
             artifact['name'] == 'rtd-wheel')[
            'archive_download_url']
    artifact_bin = io.BytesIO(
        requests.get(archive_download_url, headers={'Authorization': f'token {github_token}'}, stream=True).content)

    with zipfile.ZipFile(artifact_bin) as zf, tempfile.TemporaryDirectory() as tmpdir:
        assert len(zf.namelist()) == 1
        zf.extractall(tmpdir)
        subprocess.check_call(
            [sys.executable, '-m', 'pip', 'install', '--force-reinstall', tmpdir + '/' + zf.namelist()[0]])
html_css_files = ["readthedocs-custom.css"]  # Override some CSS settings

import egttools

# The short X.Y version
version = '.'.join(egttools.__version__.split('.')[:2])
# The full version, including alpha/beta/rc tags
release = egttools.__version__

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
#
# This is also used if you do content translation via gettext catalogs.
# Usually you set "language" from the command line for these cases.
language = 'en'

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ['_build', 'Thumbs.db', '.DS_Store', '**.ipynb_checkpoints']

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = 'sphinx'

# If true, `todo` and `todoList` produce output, else they produce nothing.
todo_include_todos = True

# enable imported members
autosummary_imported_members = True
# autosummary_implicit_namespaces = True
# Autodoc configuration
autodoc_member_order = 'groupwise'

# Intersphinx configuration
intersphinx_mapping = {'python': ('https://docs.python.org/3', None),
                       'numpy': ('https://numpy.org/doc/stable/', None),
                       'scipy': ('https://numpy.org/doc/stable/', None),
                       'networkx': ('https://networkx.org/documentation/stable/', None)
                       }

autosummary_generate = True  # Turn on sphinx.ext.autosummary
autoclass_content = "both"  # Add __init__ doc (ie. params) to class summaries
html_show_sourcelink = False  # Remove 'view source code' from top of page (for html, not python)
autodoc_inherit_docstrings = True  # If no docstring, inherit from base class
set_type_checking_flag = True  # Enable 'expensive' imports for sphinx_autodoc_typehints
nbsphinx_allow_errors = True  # Continue through Jupyter errors
# autodoc_typehints = "description" # Sphinx-native method. Not as good as sphinx_autodoc_typehints
add_module_names = False  # Remove namespaces from class/method signatures

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

default_role = 'py:obj'
nitpicky = True
nitpick_ignore = [('py:class', 'pybind11_builtins.pybind11_object'),
                  ('py:class', 'List'),
                  ('py:class', 'Positive'),
                  ('py:class', 'NonNegative'),
                  ('py:class', 'numpy.uint64'),
                  ('py:class', 'numpy.int64'),
                  ('py:class', 'numpy.float64'),
                  ('py:class', 'numpy.complex128'),
                  ('py:obj', 'List'),
                  ('py:class', 'm'),
                  ('py:class', 'n'),
                  ('py:class', '1')]

# Links to ignore
linkcheck_ignore = [
    'https://doi.org/10.1098/rspb.2008.1126',
    'https://royalsocietypublishing.org/doi/10.1098/rspb.2008.1126',
    'https://doi.org/10.1073/pnas.1015648108',
    'https://pnas.org/doi/full/10.1073/pnas.1015648108',
    'https://doi.org/10.1073/pnas.0709546105',
    'https://pnas.org/doi/full/10.1073/pnas.0709546105',
]

if on_rtd:
    branch_or_tag = branch or 'v{}'.format(release)
else:
    rev_parse_name = subprocess.check_output(['git', 'rev-parse', '--abbrev-ref', 'HEAD']).decode('ascii').strip()
    branch_or_tag = rev_parse_name if rev_parse_name != 'HEAD' else 'v{}'.format(release)

# nbsphinx_prolog = """
# {{% set docname = 'docs/' + env.doc2path(env.docname, base=False) %}}
#
# .. only:: html
#
# """
# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = 'default' if on_rtd else 'sphinx_rtd_theme'
html_logo = "images/logo.png"

# Theme options are theme-specific and customize the look and feel of a theme
# further.  For a list of options available for each theme, see the
# documentation.
#
# html_theme_options = {}

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ['_static']

# Custom sidebar templates, must be a dictionary that maps document names
# to template names.
#
# The default sidebars (for documents that don't match any pattern) are
# defined by theme itself.  Builtin themes are using these templates by
# default: ``['localtoc.html', 'relations.html', 'sourcelink.html',
# 'searchbox.html']``.
#
# html_sidebars = {}


# -- Options for HTMLHelp output ---------------------------------------------

# Output file base name for HTML help builder.
htmlhelp_basename = 'EGTtoolsdoc'

# -- Options for LaTeX output ------------------------------------------------

latex_elements = {
    # The paper size ('letterpaper' or 'a4paper').
    #
    # 'papersize': 'letterpaper',

    # The font size ('10pt', '11pt' or '12pt').
    #
    # 'pointsize': '10pt',

    # Additional stuff for the LaTeX preamble.
    #
    # 'preamble': '',

    # Latex figure (float) alignment
    #
    # 'figure_align': 'htbp',
}

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title,
#  author, documentclass [howto, manual, or own class]).
latex_documents = [
    (master_doc, 'EGTtools.tex', 'EGTtools Documentation',
     'Elias Fernández', 'manual'),
]

# -- Options for manual page output ------------------------------------------

# One entry per manual page. List of tuples
# (source start file, name, description, authors, manual section).
man_pages = [
    (master_doc, 'egttools', 'EGTtools Documentation',
     [author], 1)
]

# -- Options for Texinfo output ----------------------------------------------

# Grouping the document tree into Texinfo files. List of tuples
# (source start file, target name, title, author,
#  dir menu entry, description, category)
texinfo_documents = [
    (master_doc, 'EGTtools', 'EGTtools Documentation',
     author, 'EGTtools', 'One line description of project.',
     'Miscellaneous'),
]

latex_logo = "images/logo-full.pdf"

# -- Options for Epub output -------------------------------------------------

# Bibliographic Dublin Core info.
epub_title = project
epub_author = author
epub_publisher = author
epub_copyright = copyright

# The unique identifier of the text. This can be a ISBN number
# or the project homepage.
#
# epub_identifier = ''

# A unique identification for the text.
#
# epub_uid = ''

# A list of files that should not be packed into the epub file.
epub_exclude_files = ['search.html']

# -- Extension configuration -------------------------------------------------

# -- Options for intersphinx extension ---------------------------------------



"""
API reference documentation for `behaviors.NormalForm` submodule.
"""
try:
    from egttools.numerical.numerical_.behaviors.NormalForm import AbstractNFGStrategy
except Exception:
    raise Exception("numerical package not initialized")
else:
    import egttools.behaviors.NormalForm.TwoActions as TwoActions

__all__ = ['AbstractNFGStrategy', 'TwoActions']



import re

RE_ARGS_KWARGS = re.compile(r'(?<![*\\])\*args|(?<!\\)\*\*kwargs')


def fix_args_kwargs(app, what, name, obj, options, lines):
    for i, line in enumerate(lines):
        lines[i] = RE_ARGS_KWARGS.sub(lambda m: m.group().replace('*', '\\*'), line)


def setup(app):
    app.connect('autodoc-process-docstring', fix_args_kwargs)



# Copyright (c) 2019-2020  Elias Fernandez
#
# This file is part of EGTtools.
#
# EGTtools is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# EGTtools is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with EGTtools.  If not, see <http://www.gnu.org/licenses/>

# import os
# import sys
# sys.path.remove(os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))
#
# print(sys.path)
# import egttools



# Copyright (c) 2019-2021  Elias Fernandez
#
# This file is part of EGTtools.
#
# EGTtools is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# EGTtools is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with EGTtools.  If not, see <http://www.gnu.org/licenses/>

from typing import Dict, Tuple, Union

from .. import AbstractNFGStrategy
from egttools import Random

import numpy as np

__all__ = ['EpsilonTFT', 'EpsilonGRIM', 'Detective', 'MemoryOneStrategy']


class EpsilonTFT(AbstractNFGStrategy):
    def __init__(self, p: float, epsilon: float):
        """
        A TFT player with randomized first action and probability
        of making mistakes.

        This player acts exactly as Tit-for-Tat (repeats
        the last action of the opponent), however in the
        first round it will cooperate with probability :param p
        and in the subsequent rounds it has a probability
        :param epsilon of making a mistake and changing its action.

        Parameters
        ----------
        p : float
            Probability of cooperating in the first round
        epsilon : float
            Probability of making a mistake in any round after round 1.
        """
        super().__init__()
        self.p_ = p
        self.epsilon_ = epsilon
        self.random_device = np.random.default_rng(Random.generate())
        self.is_stochastic_strategy = True

    def get_action(self, time_step: int, action_prev: int):
        if time_step == 0:
            return 0 if self.random_device.random() < self.p_ else 1
        else:
            return action_prev if self.random_device.random() >= self.epsilon_ else (action_prev + 1) % 2

    def type(self):
        return "NFGStrategies::EpsilonTFT"

    def is_stochastic(self):
        return self.is_stochastic_strategy

    def __str__(self):
        return self.type()


class EpsilonGRIM(AbstractNFGStrategy):
    def __init__(self, p: float, epsilon: float):
        """
        A GRIM player with randomized first action and probability
        of making mistakes.

        This player acts exactly as GRIM (cooperates until the opponent defects),
        however in the first round it will cooperate with probability :param p
        and in the subsequent rounds it has a probability
        :param epsilon of making a mistake and changing its action.

        Parameters
        ----------
        p : float
            Probability of cooperating in the first round
        epsilon : float
            Probability of making a mistake in any round after round 1.
        """
        super().__init__()
        self.p_ = p
        self.epsilon_ = epsilon
        self.random_device = np.random.default_rng(Random.generate())
        self.action_ = 1
        self.is_stochastic_strategy = True

    def get_action(self, time_step: int, action_prev: int):
        if time_step == 0:
            self.action_ = 1  # reset action to cooperation
            return 0 if self.random_device.random() < self.p_ else 1

        if self.action_ == 1:
            if action_prev == 0:
                self.action_ = 0

        action = self.action_

        if self.random_device.random() < self.epsilon_:
            # In this case we reset the trigger
            self.action_ = 1
            # And the real action in this round is the opposite of the original intention
            action = (action + 1) % 2

        return action

    def type(self):
        return "NFGStrategies::EpsilonGRIM"

    def is_stochastic(self):
        return self.is_stochastic_strategy

    def __str__(self):
        return self.type()


class Detective(AbstractNFGStrategy):
    def __init__(self):
        """
        A Detective player who tries to analyze the opponent.

        This player will always play the same initial sequence of
        Cooperate, Defect, Cooperate, Cooperate. If the opponent defects
        during this initial sequence, then Defective will play TFT from the 5th
        round on. Otherwise, Detective will play always Defect.
        """
        super().__init__()
        self.initial_sequence_ = [1, 0, 1, 1]
        self.cheated_ = False
        self.is_stochastic_strategy = False

    def get_action(self, time_step: int, action_prev: int):
        if time_step == 0:  # reset internal variables
            self.cheated_ = False
        if time_step < 4:
            if time_step > 0 and action_prev == 0:
                self.cheated_ = True
            return self.initial_sequence_[time_step]
        else:
            if self.cheated_:
                return action_prev
            else:
                return 0

    def type(self):
        return "NFGStrategies::Detective"

    def is_stochastic(self):
        return self.is_stochastic_strategy

    def __str__(self):
        return self.type()


class MemoryOneStrategy(AbstractNFGStrategy):
    def __init__(self, action_first_round: Union[int, float],
                 strategy: Union[Dict[Tuple[int, int], int], Dict[Tuple[int, int], float]], is_stochastic: bool):
        """
        Defines a Memory One strategy.

        Parameters
        ----------
        action_first_round: Union[int, float]
            Indicates the action this strategy will play in the first round. In the case that `is_stochastic`
            is True, then this value should be a probability of Cooperation
        strategy: Union[Dict[Tuple[int, int], int], Dict[Tuple[int, int], float]]
            A dictionary with tuples defining the action/probability of cooperation for
            each pair of previous actions of self and the opponent, e.g., CC, DC....
        is_stochastic: bool
            Indicates whether the strategy is stochastic or not. If it is stochastic, then the values both
            for the `action_first_round` and the strategy should be probabilities of Cooperation. If it is
            False, then 1 - indicates Cooperation and 0 - indicates Defection.
        """
        super().__init__()
        self.action_first_round_ = action_first_round
        self.strategy_ = strategy
        self.action_prev_self_ = action_first_round
        self.random_device = np.random.default_rng(Random.generate())
        self.is_stochastic_strategy = is_stochastic

    def get_action(self, time_step: int, action_prev: int):
        if time_step == 0:
            if not self.is_stochastic_strategy:
                self.action_prev_self_ = self.action_first_round_
            else:
                self.action_prev_self_ = 1 if self.random_device.random() < self.strategy_[
                    (self.action_prev_self_, action_prev)] else 0
            return self.action_first_round_
        else:
            if not self.is_stochastic_strategy:
                action = self.strategy_[(self.action_prev_self_, action_prev)]
            else:
                action = 1 if self.random_device.random() < self.strategy_[(self.action_prev_self_, action_prev)] else 0
            self.action_prev_self_ = action
            return action

    def type(self):
        return "NFGStrategies::MemoryOneStrategy"

    def is_stochastic(self):
        return self.is_stochastic_strategy

    def __str__(self):
        return self.type()



# Copyright (c) 2019-2021  Elias Fernandez
#
# This file is part of EGTtools.
#
# EGTtools is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# EGTtools is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with EGTtools.  If not, see <http://www.gnu.org/licenses/>

from typing import List, Union
import numpy as np

from .. import (sample_simplex, )
from . import AbstractNPlayerGame
from ..behaviors.pgg_behaviors import PGGOneShotStrategy


class PGG(AbstractNPlayerGame):
    def __init__(self, group_size: int, cost: float, multiplying_factor: float,
                 strategies: List[PGGOneShotStrategy]) -> None:
        """
        Classical Public Goods game with only 2 possible contributions (o or cost).

        Parameters
        ----------
        group_size: int
            Size of the group playing the game.
        cost: float
            Cost of cooperation.
        multiplying_factor: float
            The sum of contributions to the public good is multiplied by this factor before being divided equally
            among all players.
        strategies: List[egttools.behaviors.pgg_behaviors.PGGOneShotStrategy]
            A list of strategies that will play the game.
        """
        AbstractNPlayerGame.__init__(self, len(strategies), group_size)
        self.group_size_ = group_size
        self.c_ = cost
        self.r_ = multiplying_factor
        self.strategies_ = strategies
        self.nb_strategies_ = len(strategies)

        self.nb_group_configurations_ = self.nb_group_configurations()
        self.calculate_payoffs()

    def play(self, group_composition: Union[List[int], np.ndarray], game_payoffs: np.ndarray) -> None:
        # Gather contributions
        contributions = 0.0
        non_zero = []
        for i, strategy_count in enumerate(group_composition):
            if strategy_count == 0:
                continue
            else:
                non_zero.append(i)
                action = self.strategies_[i].get_action()
                if action == 1:
                    contributions += strategy_count * self.c_
                    game_payoffs[i] = - self.c_

        benefit = (contributions * self.r_) / self.group_size_
        game_payoffs[non_zero] += benefit

    def calculate_payoffs(self) -> np.ndarray:
        payoffs_container = np.zeros(shape=(self.nb_strategies_,), dtype=np.float64)
        for i in range(self.nb_group_configurations_):
            # Get group composition
            group_composition = sample_simplex(i, self.group_size_, self.nb_strategies_)
            self.play(group_composition, payoffs_container)
            for strategy_index, strategy_payoff in enumerate(payoffs_container):
                self.update_payoff(strategy_index, i, strategy_payoff)
            # Reinitialize payoff vector
            payoffs_container[:] = 0

        return self.payoffs()

    def __str__(self) -> str:
        string = f'''
        Python implementation of a public goods game.\n
        Game parameters
        -------
        cost = {self.c_}\n
        multiplication_factor = {self.r_}\n
        Strategies
        -------
        nb_strategies = {self.nb_strategies_}\n
        strategies = {self.strategies_}\n
        '''
        return string

    def type(self) -> str:
        return "PGG"

    def save_payoffs(self, file_name: str) -> None:
        with open(file_name, 'w') as f:
            f.write('Payoffs for each type of player and each possible state:\n')
            f.write(f'rows: {" ,".join([strategy.type for strategy in self.strategies_])}\n')
            f.write('cols: all possible group compositions starting at (0, 0, ..., group_size)\n')
            f.write(f'{self.payoffs_}')
            f.write(f'group_size = {self.group_size_}\n')
            f.write(f'cost = {self.c_}\n')
            f.write(f'multiplying_factor = {self.r_}\n')



import numpy as np
from .abstract_cpr_strategy import AbstractCPRStrategy


def high_extraction(a: float, b: float, group_size: int) -> float:
    return a / (b * group_size)


def fair_extraction(a: float, b: float, group_size: int) -> float:
    group_optimal = a / (2 * b)
    return group_optimal / group_size


def nash_extraction(a: float, b: float, group_size: int) -> float:
    group_max = (group_size / (group_size + 1)) * (a / b)
    return group_max / group_size


def payoff_no_commitment(a: float, b: float, extraction: float, group_extraction: float) -> float:
    return (extraction / group_extraction) * ((a * group_extraction) - (b * np.power(group_extraction, 2)))


class FixedExtraction(AbstractCPRStrategy):

    def __init__(self, extraction: float, accepts_commitment: bool):
        self.extraction = extraction
        self.accepts_commitment = accepts_commitment

    def get_extraction(self, a: float, b: float, group_size: int, commitment: bool = False) -> float:
        return self.extraction

    @staticmethod
    def get_payoff(a: float, b: float, extraction: float, group_extraction: float, fine: float = 0,
                   cost: float = 0, commitment: bool = False) -> float:
        return payoff_no_commitment(a, b, extraction, group_extraction)

    def would_like_to_commit(self) -> bool:
        return self.accepts_commitment

    def proposes_commitment(self) -> bool:
        return False

    def is_commitment_validated(self, nb_committers: int) -> bool:
        return False

    def type(self) -> str:
        return f"EXT{self.extraction}{self.accepts_commitment}"

    def __str__(self) -> str:
        return f"FixedExtraction_e{self.extraction}_c{self.accepts_commitment}"


class FairExtraction(AbstractCPRStrategy):

    def get_extraction(self, a: float, b: float, group_size: int, commitment: bool = False) -> float:
        return fair_extraction(a, b, group_size)

    @staticmethod
    def get_payoff(a: float, b: float, extraction: float, group_extraction: float, fine: float = 0,
                   cost: float = 0, commitment: bool = False) -> float:
        return payoff_no_commitment(a, b, extraction, group_extraction)

    def would_like_to_commit(self) -> bool:
        return True

    def proposes_commitment(self) -> bool:
        return False

    def is_commitment_validated(self, nb_committers: int) -> bool:
        if nb_committers > 0:
            return True
        else:
            return False

    def type(self) -> str:
        return "FAIR"

    def __str__(self) -> str:
        return "FairExtraction"


class HighExtraction(AbstractCPRStrategy):

    def get_extraction(self, a: float, b: float, group_size: int, commitment: bool = False) -> float:
        return high_extraction(a, b, group_size)

    @staticmethod
    def get_payoff(a: float, b: float, extraction: float, group_extraction: float, fine: float = 0,
                   cost: float = 0, commitment: bool = False) -> float:
        return payoff_no_commitment(a, b, extraction, group_extraction)

    def would_like_to_commit(self) -> bool:
        return False

    def proposes_commitment(self) -> bool:
        return False

    def is_commitment_validated(self, nb_committers: int) -> bool:
        return False

    def type(self) -> str:
        return "HIGH"

    def __str__(self) -> str:
        return "HighExtraction"


class NashExtraction(AbstractCPRStrategy):

    def get_extraction(self, a: float, b: float, group_size: int, commitment: bool = False) -> float:
        return nash_extraction(a, b, group_size)

    @staticmethod
    def get_payoff(a: float, b: float, extraction: float, group_extraction: float, fine: float = 0,
                   cost: float = 0, commitment: bool = False) -> float:
        return payoff_no_commitment(a, b, extraction, group_extraction)

    def would_like_to_commit(self) -> bool:
        return False

    def proposes_commitment(self) -> bool:
        return False

    def is_commitment_validated(self, nb_committers: int) -> bool:
        return False

    def type(self) -> str:
        return "NASH"

    def __str__(self) -> str:
        return "NashExtraction"


class CommitmentStrategy(AbstractCPRStrategy):
    def __init__(self, commitment_threshold):
        self.commitment_threshold = commitment_threshold

    def get_extraction(self, a: float, b: float, group_size: int, commitment: bool = False) -> float:
        if commitment:  # extracts Fair
            return fair_extraction(a, b, group_size)
        else:  # extracts High
            return high_extraction(a, b, group_size)

    @staticmethod
    def get_payoff(a: float, b: float, extraction: float, group_extraction: float, fine: float = 0,
                   cost: float = 0, commitment: bool = False) -> float:
        if commitment:
            return payoff_no_commitment(a, b, extraction, group_extraction) - cost
        else:
            return payoff_no_commitment(a, b, extraction, group_extraction)

    def would_like_to_commit(self) -> bool:
        return True

    def proposes_commitment(self) -> bool:
        return True

    def is_commitment_validated(self, nb_committers: int) -> bool:
        return False if nb_committers < self.commitment_threshold else True

    def type(self) -> str:
        return f"COM{self.commitment_threshold}"

    def __str__(self) -> str:
        return f"CommitmentStrategy{self.commitment_threshold}"


class FakeStrategy(AbstractCPRStrategy):

    def get_extraction(self, a: float, b: float, group_size: int, commitment: bool = False) -> float:
        return high_extraction(a, b, group_size)

    @staticmethod
    def get_payoff(a: float, b: float, extraction: float, group_extraction: float, fine: float = 0,
                   cost: float = 0, commitment: bool = False) -> float:
        if commitment:
            return payoff_no_commitment(a, b, extraction, group_extraction) - fine
        else:
            return payoff_no_commitment(a, b, extraction, group_extraction)

    def would_like_to_commit(self) -> bool:
        return True

    def proposes_commitment(self) -> bool:
        return False

    def is_commitment_validated(self, nb_committers: int) -> bool:
        return True if nb_committers > 0 else False

    def type(self) -> str:
        return "FAKE"

    def __str__(self) -> str:
        return "FakeStrategy"


class FreeStrategy(AbstractCPRStrategy):

    def get_extraction(self, a: float, b: float, group_size: int, commitment: bool = False) -> float:
        if commitment:
            return fair_extraction(a, b, group_size)
        else:
            return high_extraction(a, b, group_size)

    @staticmethod
    def get_payoff(a: float, b: float, extraction: float, group_extraction: float, fine: float = 0,
                   cost: float = 0, commitment: bool = False) -> float:
        return payoff_no_commitment(a, b, extraction, group_extraction)

    def would_like_to_commit(self) -> bool:
        return True

    def proposes_commitment(self) -> bool:
        return False

    def is_commitment_validated(self, nb_committers: int) -> bool:
        return True if nb_committers > 0 else False

    def type(self) -> str:
        return "FREE"

    def __str__(self) -> str:
        return "FreeStrategy"



"""The structure submodule contains population structures"""

try:
    from ..numerical_.structure import (AbstractStructure, Network, NetworkGroup,
                                        NetworkSync, NetworkGroupSync)
except Exception:
    raise Exception("numerical package not initialized")

__all__ = ['AbstractStructure', 'Network', 'NetworkGroup', 'NetworkSync', 'NetworkGroupSync']



from egttools.distributions import multinomial_pmf
from egttools import sample_simplex, calculate_nb_states
from scipy.stats import multinomial
import numpy as np


def test_multinomial_two_strategies():
    group_size = 10
    nb_strategies = 2
    nb_group_combinations = calculate_nb_states(group_size, nb_strategies)

    nb_points = 101
    strategy_i = np.linspace(0, 1, num=nb_points, dtype=np.float64)
    states = [np.array([i, 1 - i]) for i in strategy_i]

    for state in states:
        for group_index in range(nb_group_combinations):
            group = sample_simplex(group_index, group_size, nb_strategies)
            prob_scipy = multinomial.pmf(group, group_size, state)
            prob_egttools = multinomial_pmf(group, group_size, state)

            assert np.isclose(prob_egttools, prob_scipy)



# Copyright (c) 2019-2021  Elias Fernandez
#
# This file is part of EGTtools.
#
# EGTtools is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# EGTtools is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with EGTtools.  If not, see <http://www.gnu.org/licenses/>

from typing import List
import numpy as np
from scipy.stats import multivariate_hypergeom

from . import AbstractGame
from .. import calculate_nb_states, calculate_state, sample_simplex


class InformalRiskGame(AbstractGame):
    def __init__(self, group_size: int, cost: float, multiplying_factor: float, strategies: List) -> None:
        """
        Game of informal risk sharing.

        This game has been taken from the model introduced in

        ```Santos, F. P., Pacheco, J. M., Santos, F. C., & Levin, S. A. (2021).
        Dynamics of informal risk sharing in collective index insurance.
        Nature Sustainability. https://doi.org/10.1038/s41893-020-00667-2```

        to investigate the dynamics of a collective index insurance with informal risk sharing.

        Parameters
        ----------
        group_size
        cost
        multiplying_factor
        strategies
        """
        AbstractGame.__init__(self)
        self.group_size_ = group_size
        self.c_ = cost
        self.r_ = multiplying_factor
        self.strategies_ = strategies
        self.nb_strategies_ = len(strategies)
        self.nb_states_ = calculate_nb_states(self.group_size_, self.nb_strategies_)
        self.payoffs_ = np.zeros(shape=(self.nb_strategies_, self.nb_states_), dtype=np.float64)
        self.calculate_payoffs()

    def play(self, group_composiiton: List[int], game_payoffs: np.ndarray) -> None:
        # Gather contributions
        contributions = 0.0
        non_zero = []
        for i, strategy_count in enumerate(group_composiiton):
            if strategy_count == 0:
                continue
            else:
                non_zero.append(i)
                action = self.strategies_[i].get_action()
                if action == 1:
                    contributions += strategy_count * self.c_
                    game_payoffs[i] = - self.c_

        benefit = (contributions * self.r_) / self.group_size_
        game_payoffs[non_zero] += benefit

    def calculate_payoffs(self) -> np.ndarray:
        payoffs_container = np.zeros(shape=(self.nb_strategies_,), dtype=np.float64)
        for i in range(self.nb_states_):
            # Get group composition
            group_composition = sample_simplex(i, self.group_size_, self.nb_strategies_)
            self.play(group_composition, payoffs_container)
            for strategy_index, strategy_payoff in enumerate(payoffs_container):
                self.payoffs_[strategy_index, i] = strategy_payoff
            # Reinitialize payoff vector
            payoffs_container[:] = 0

        return self.payoffs_

    def calculate_fitness(self, player_strategy: int, pop_size: int,
                          population_state: np.ndarray) -> float:
        """
        Calculates the Fitness of an strategy for a given population state.

        The calculation is done by computing the expected payoff over all possible group combinations
        for the given population state: $ fitness = \\sum_{states} payoff * P(state) $
        Parameters
        ----------
        player_strategy : index of the strategy
        pop_size : size of the population - Only necessary for compatibility with the C++ implementation
                 (might be eliminated in the future)
        population_state : vector with the population state (the number of players adopting each strategy)

        Returns
        -------
        The fitness of the population.
        """
        # multivariate PDF
        population_state[player_strategy] -= 1
        rv = multivariate_hypergeom(population_state, self.group_size_ - 1)
        population_state[player_strategy] += 1

        fitness = 0.0
        # Iterate over all possible group compositions
        for i in range(self.nb_states_):
            group_composition = sample_simplex(i, self.group_size_ - 1, self.nb_strategies_)
            # Estimate probability of the current group composition
            if group_composition[player_strategy] > 0:
                group_composition[player_strategy] -= 1
                fitness += self.payoffs[player_strategy, i] * rv.pmf(x=group_composition)

        return fitness

    def __str__(self) -> str:
        string = f'''
        Python implementation of a public goods game.\n
        Game parameters
        -------
        cost = {self.c_}\n
        multiplication_factor = {self.r_}\n
        Strategies
        -------
        nb_strategies = {self.nb_strategies_}\n
        strategies = {self.strategies_}\n
        '''
        return string

    @property
    def nb_strategies(self) -> int:
        return self.nb_strategies_

    @property
    def type(self) -> str:
        return "PGG"

    @property
    def payoffs(self) -> np.ndarray:
        return self.payoffs_

    def payoff(self, strategy: int, group_composition: List[np.uint64]) -> float:
        if strategy > self.nb_strategies_:
            raise IndexError(f'You must specify a valid index for the strategy [0, {self.nb_strategies_}].')
        elif len(group_composition) != self.nb_strategies_:
            raise Exception(f'The group composition list must be of size {self.nb_strategies_}')

        return self.payoffs_[strategy, calculate_state(self.group_size_, group_composition)]

    def save_payoffs(self, file_name: str) -> None:
        with open(file_name, 'w') as f:
            f.write('Payoffs for each type of player and each possible state:\n')
            f.write(f'rows: {" ,".join([strategy.type() for strategy in self.strategies_])}\n')
            f.write('cols: all possible group compositions starting at (0, 0, ..., group_size)\n')
            f.write(f'{self.payoffs_}')
            f.write(f'group_size = {self.group_size_}\n')
            f.write(f'cost = {self.c_}\n')
            f.write(f'multiplying_factor = {self.r_}\n')



from typing import List

from . import AbstractCRDStrategy


class TimeBasedCRDStrategy(AbstractCRDStrategy):
    def __init__(self, actions_per_round: List[int]):
        """
        A CRD strategy which adapts in function of a moving average of the contributions of the rest of the group.

        Parameters
        ----------
        actions_per_round: List[int]
            Defines the action which this strategy will play in every round. If you change the number of rounds
            of the game, you should redefine this list, otherwise there will be an exception!!
        """
        super().__init__()
        self.actions_per_round = actions_per_round

    def get_action(self, time_step: int, group_contributions_prev: int):
        return self.actions_per_round[time_step]

    def type(self):
        return "TimeBasedCRDStrategy({})".format(self.actions_per_round)

    def __str__(self):
        return self.type()



from typing import Tuple

import numpy as np
import pytest

egttools = pytest.importorskip("egttools")

replicator_equation = egttools.analytical.replicator_equation
find_saddle_type_and_gradient_direction = egttools.utils.find_saddle_type_and_gradient_direction


@pytest.fixture
def setup_hawk_dove_gradients_and_saddle_points() -> Tuple[np.array, np.array]:
    nb_points = 101
    strategy_i = np.linspace(0, 1, num=nb_points, dtype=np.float64)
    strategy_j = 1 - strategy_i
    states = np.array((strategy_i, strategy_j)).T

    # Payoff matrix
    v, d, t = 2, 3, 1
    payoffs = np.array([
        [(v - d) / 2, v],
        [0, (v / 2) - t],
    ])

    # Calculate gradient
    gradients = np.array([replicator_equation(states[i], payoffs)[0] for i in range(len(states))])
    epsilon = 1e-3
    saddle_points_idx = np.where((gradients <= epsilon) & (gradients >= -epsilon))[0]

    return gradients, saddle_points_idx


def test_find_saddle_type_and_gradient_direction(setup_hawk_dove_gradients_and_saddle_points) -> None:
    """
    This test is used to check if the types of saddle points are correctly
    identified and if the arrow directions are correct.
    """

    saddle_type, gradient_direction = find_saddle_type_and_gradient_direction(
        *setup_hawk_dove_gradients_and_saddle_points)

    assert not saddle_type[0]  # First
    assert not saddle_type[2]  # Last saddle points are not stable
    assert saddle_type[1]  # Middle saddle point is stable

    # check that there are only 2 arrows
    assert gradient_direction.shape == (2, 2)
    # check tha the first arrows goes from 0 to the stable point
    assert tuple(gradient_direction[0]) == (0., 0.79)
    # check tha the second goes from 1 to the stable point
    assert tuple(gradient_direction[1]) == (1., 0.81)



# Copyright (c) 2019-2021  Elias Fernandez
#
# This file is part of EGTtools.
#
# EGTtools is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# EGTtools is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with EGTtools.  If not, see <http://www.gnu.org/licenses/>

"""
Set of vectorized functions that can be used to apply these functions on large tensors.
"""
try:
    from ..numerical.numerical_ import \
        vectorized_replicator_equation_n_player as cpp_vectorized_replicator_equation_n_player
except Exception:
    raise Exception("numerical package not initialized")
else:
    import numpy as np
    from ..plotting.helpers import barycentric_to_xy_coordinates
    from ..analytical import replicator_equation_n_player


def vectorized_replicator_equation(frequencies: np.ndarray, payoffs: np.ndarray) -> np.ndarray:
    """
    This function provides an easy way to calculate a matrix of gradients in a simplex in one go.

    The input `frequencies` is expected to be a 3 dimensional tensor of shape (p, m, n) while the payoffs
    matrix is expected to be of shape (p, p).

    The main intention of this helper function is to facilitate
    the calculation of the gradients that are required by the `plot_gradients` method of the
    `egttools.Simplex2D` class.

    Parameters
    ----------
    frequencies: numpy.ndarray[p,m,n]
        A 3 dimensional tensor containing the set of population frequencies for which the gradient should be
        calculated.
    payoffs: numpy.ndarray[p,p]
        A 2 dimensional matrix containing the payoffs of the game.

    Returns
    -------
    numpy.ndarray[p,m,n]
        The gradients for each of the input frequencies.
    """
    axs = [np.dot(payoffs, frequencies[:, i, :]) for i in range(frequencies.shape[1])]
    return np.asarray([frequencies[:, i, :] * (axs[i] - np.diagonal(np.dot(frequencies[:, i, :].T, axs[i]))) for i in
                       range(frequencies.shape[1])]).swapaxes(0, 1)


def vectorized_replicator_equation_n_player(frequencies: np.ndarray, payoffs: np.ndarray,
                                            group_size: int) -> np.ndarray:
    """
    This function provides an easy way to calculate a matrix of gradients in a simplex in one go.

    The input `frequencies` is expected to be a 3 dimensional tensor of shape (p, m, n) while the payoffs
    matrix is expected to be of shape (p, p).

    The main intention of this helper function is to facilitate
    the calculation of the gradients that are required by the `plot_gradients` method of the
    `egttools.Simplex2D` class.

    Parameters
    ----------
    frequencies: numpy.ndarray[p,m,n]
        A 3 dimensional tensor containing the set of population frequencies for which the gradient should be
        calculated.
    payoffs: numpy.ndarray[p,p]
        A 2 dimensional matrix containing the payoffs of the game.
    group_size: int
        Size of the group

    Returns
    -------
    numpy.ndarray[p,m,n]
        The gradients for each of the input frequencies.
    """
    res1, res2, res3 = cpp_vectorized_replicator_equation_n_player(frequencies[0, :, :], frequencies[1, :, :],
                                                                   frequencies[2, :, :], payoffs, group_size)

    results = np.zeros_like(frequencies)
    results[0, :, :] = res1[:, :]
    results[1, :, :] = res2[:, :]
    results[2, :, :] = res3[:, :]

    return results


def vectorized_barycentric_to_xy_coordinates(barycentric_coordinates: np.ndarray, corners: np.ndarray) -> np.ndarray:
    """
    Transform a tensor of barycentric coordinates to cartesian coordinates.

    Parameters
    ----------
    barycentric_coordinates : numpy.ndarray[3,m,n]
        Expects a matrix in which the first dimension corresponds to the vector of 3-demensional barycentric
        coordinates.

    corners : numpy.ndarray[3,]
        The corners of the triangle

    Returns
    -------
    numpy.ndarray[2,m,n]
        The tensor of cartesian coordinates.
    """
    return np.asarray([[barycentric_to_xy_coordinates(barycentric_coordinates[:, i, j], corners)
                        for j in range(barycentric_coordinates.shape[1])]
                       for i in range(barycentric_coordinates.shape[2])])



# Copyright (c) 2019-2022  Elias Fernandez
#
# This file is part of EGTtools.
#
# EGTtools is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# EGTtools is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with EGTtools.  If not, see <http://www.gnu.org/licenses/>

"""Simplified plotting functions"""
import matplotlib.pyplot as plt
import numpy as np

from typing import Optional, Tuple, Callable, List
from ..games import (AbstractGame, Matrix2PlayerGameHolder, MatrixNPlayerGameHolder, )
from .. import (calculate_nb_states, )
from .helpers import (barycentric_to_xy_coordinates,
                      xy_to_barycentric_coordinates, calculate_stability,
                      find_roots_in_discrete_barycentric_coordinates)
from ..analytical import (replicator_equation, PairwiseComparison, )
from ..analytical import replicator_equation_n_player
from ..analytical.utils import check_if_there_is_random_drift, check_replicator_stability_pairwise_games, find_roots
from ..helpers.vectorized import (vectorized_replicator_equation, vectorized_replicator_equation_n_player,
                                  vectorized_barycentric_to_xy_coordinates)
from . import Simplex2D


def plot_replicator_dynamics_in_simplex(payoff_matrix: np.ndarray,
                                        group_size: int = 2,
                                        nb_points_simplex: int = 100,
                                        nb_of_initial_points_for_root_search: int = 10,
                                        atol: float = 1e-7,
                                        atol_equal: float = 1e-12,
                                        method_find_roots: str = 'hybr',
                                        atol_stability_pos: float = 1e-4, atol_stability_neg: float = 1e-4,
                                        atol_stability_zero: float = 1e-4,
                                        figsize: Tuple[int, int] = (10, 8),
                                        ax: Optional[plt.axis] = None) -> Tuple[Simplex2D,
                                                                                Callable[[np.ndarray, int], np.ndarray],
                                                                                List[np.ndarray],
                                                                                List[np.ndarray],
                                                                                List[int]]:
    """
    Helper function to simplify the plotting of the replicator dynamics in a 2 Simplex.

    Parameters
    ----------
    payoff_matrix: numpy.ndarray
        The square payoff matrix. Group games are still unsupported in the replicator dynamics. This feature will
        soon be added.
    group_size: int
        size of the group
    nb_points_simplex: int
        Number of initial points to draw in the simplex
    nb_of_initial_points_for_root_search: int
        Number of initial points used in the method that searches for the roots of the replicator equation
    atol: float
        Tolerance to consider a value equal to zero. This is used to check if an edge has random drift. By default,
        the tolerance is 1e-7.
    atol_equal: float
        Tolerance to consider two arrays equal.
    method_find_roots: str
        Method used in scipy.optimize.root
    atol_stability_neg: float
        Tolerance used to determine the stability of the roots. This is used to determine whether an
        eigenvalue is negative.
    atol_stability_pos: float
        Tolerance used to determine the stability of the roots. This is used to determine whether an
        eigenvalue is positive.
    atol_stability_zero: float
        Tolerance used to determine the stability of the roots. This is used to determine whether an
        eigenvalue is zero.
    figsize: Tuple[int, int]
        Size of the figure. This parameter is only used if the ax parameter is not defined.
    ax: Optional[matplotlib.pyplot.axis]
        A matplotlib figure axis.

    Returns
    -------
    Tuple[Simplex2D, Callable[[numpy.ndarray, int], numpy.ndarray], List[numpy.ndarray], List[numpy.ndarray], List[int]]

    A tuple with the simplex object which can be used to add more features to the plot, the function that
    can be used to calculate gradients and should be passed to `Simplex2D.draw_trajectory_from_roots` and
    `Simplex2D.draw_scatter_shadow`, a list of the roots in barycentric coordinates, a list of the roots in
    cartesian coordinates and a list of booleans or integers indicating whether the roots are stable.

    """
    if (group_size > 2) and (payoff_matrix.shape[1] == payoff_matrix.shape[0]):
        nb_group_configurations = calculate_nb_states(group_size, payoff_matrix.shape[0])
        if payoff_matrix.shape[1] != nb_group_configurations:
            raise ValueError("The number of columns of the payoff matrix must be equal "
                             "the the number of possible group configurations, when group_size > 2.")

    simplex = Simplex2D(nb_points=nb_points_simplex)
    simplex.add_axis(figsize, ax)

    random_drift = check_if_there_is_random_drift(payoff_matrix, group_size=group_size, atol=atol)
    simplex.add_edges_with_random_drift(random_drift)
    v = np.asarray(xy_to_barycentric_coordinates(simplex.X, simplex.Y, simplex.corners))
    if group_size > 2:
        results = vectorized_replicator_equation_n_player(v, payoff_matrix, group_size)

        def gradient_function(u):
            return replicator_equation_n_player(u, payoff_matrix, group_size)
    else:
        results = vectorized_replicator_equation(v, payoff_matrix)

        def gradient_function(u):
            return replicator_equation(u, payoff_matrix)
    xy_results = vectorized_barycentric_to_xy_coordinates(results, simplex.corners)

    ux = xy_results[:, :, 0].astype(np.float64)
    uy = xy_results[:, :, 1].astype(np.float64)

    simplex.apply_simplex_boundaries_to_gradients(ux, uy)

    roots = find_roots(gradient_function=gradient_function,
                       nb_strategies=payoff_matrix.shape[0],
                       nb_initial_random_points=nb_of_initial_points_for_root_search,
                       atol=atol_equal, tol_close_points=atol_equal, method=method_find_roots)

    roots_xy = [barycentric_to_xy_coordinates(root, corners=simplex.corners) for root in roots]

    if group_size > 2:
        stability = calculate_stability(roots, gradient_function)
        stability = [1 if x is True else -1 for x in stability]
    else:
        stability = check_replicator_stability_pairwise_games(roots, payoff_matrix, atol_neg=atol_stability_neg,
                                                              atol_pos=atol_stability_pos,
                                                              atol_zero=atol_stability_zero)

    return simplex, lambda u, t: gradient_function(u), roots, roots_xy, stability


def plot_pairwise_comparison_rule_dynamics_in_simplex(population_size: int,
                                                      beta: float,
                                                      payoff_matrix: np.ndarray = None,
                                                      game: AbstractGame = None,
                                                      group_size: Optional[int] = 2,
                                                      atol: Optional[float] = 1e-7,
                                                      figsize: Optional[Tuple[int, int]] = (10, 8),
                                                      ax: Optional[plt.axis] = None) -> \
        Tuple[Simplex2D, Callable[[np.ndarray, int], np.ndarray], List[np.ndarray], List[np.ndarray], List[
            bool], AbstractGame, PairwiseComparison]:
    """
    Helper function to simplify the plotting of the moran dynamics in a 2 Simplex.

    Parameters
    ----------
    population_size:
        Size of the finite population.
    beta:
        Intensity of selection.
    payoff_matrix:
        The square payoff matrix. Group games are still unsupported in the replicator dynamics. This feature will
        soon be added.
    game:
        Game that should contain a set of payoff matrices
    group_size:
        Size of the group. By default, we assume that interactions are pairwise (the group size is 2).
    atol:
        Tolerance to consider a value equal to zero. This is used to check if an edge has random drift. By default
        the tolerance is 1e-7.
    figsize:
        Size of the figure. This parameter is only used if the ax parameter is not defined.
    ax:
        A matplotlib figure axis.

    Returns
    -------
    A tuple with the simplex object which can be used to add more features to the plot, the function that
    can be used to calculate gradients and should be passed to `Simplex2D.draw_trajectory_from_roots` and
    `Simplex2D.draw_scatter_shadow`, a list of the roots in barycentric coordinates, a list of the roots in
    cartesian coordinates and a list of booleans indicating whether the roots are stable. It also returns the
    game class (this is important, since a new game is created when passing a payoff matrix, and if not returned,
    a reference to the game instance will disappear, and it will produce a segmentation fault). Finally, it also returns
    a reference to the evolver object.
    """
    if (payoff_matrix is None) and (game is None):
        raise Exception("You need to define either a payoff matrix or a game.")
    elif game is None:
        if (group_size is None) or (group_size < 2):
            raise Exception("group_size not be None and must be >= 2")

        if group_size == 2:
            game = Matrix2PlayerGameHolder(payoff_matrix.shape[0], payoff_matrix)
        else:
            game = MatrixNPlayerGameHolder(payoff_matrix.shape[0], group_size, payoff_matrix)

    simplex = Simplex2D(discrete=True, size=population_size, nb_points=population_size + 1)
    simplex.add_axis(figsize, ax)

    random_drift = check_if_there_is_random_drift(payoff_matrix=game.payoffs(), population_size=population_size,
                                                  group_size=group_size, beta=beta, atol=atol)
    simplex.add_edges_with_random_drift(random_drift)

    v = np.asarray(xy_to_barycentric_coordinates(simplex.X, simplex.Y, simplex.corners))
    v_int = np.floor(v * population_size).astype(np.int64)

    evolver = PairwiseComparison(population_size=population_size, game=game)
    # evolver = StochDynamics(nb_strategies=3, payoffs=payoff_matrix, pop_size=population_size, group_size=group_size)
    result = np.zeros(shape=(v_int.shape[1], v_int.shape[2], 3))
    for i in range(v_int.shape[1]):
        for j in range(v_int.shape[2]):
            result[i, j, :] = evolver.calculate_gradient_of_selection(beta, v_int[:, i, j])

    result = result.swapaxes(0, 1).swapaxes(0, 2)
    xy_results = vectorized_barycentric_to_xy_coordinates(result, simplex.corners)

    ux = xy_results[:, :, 0].astype(np.float64)
    uy = xy_results[:, :, 1].astype(np.float64)

    simplex.apply_simplex_boundaries_to_gradients(ux, uy)

    roots = find_roots_in_discrete_barycentric_coordinates(
        lambda u: population_size * evolver.calculate_gradient_of_selection(beta, u), population_size,
        nb_interior_points=calculate_nb_states(population_size,
                                               3),
        atol=1e-1)
    roots_xy = [barycentric_to_xy_coordinates(x, simplex.corners) for x in roots]
    stability = calculate_stability(roots, lambda u: population_size * evolver.calculate_gradient_of_selection(beta, u))

    return (simplex,
            lambda u, t: population_size * evolver.calculate_gradient_of_selection(beta, u),
            roots, roots_xy, stability, game, evolver)


def plot_pairwise_comparison_rule_dynamics_in_simplex_without_roots(population_size: int,
                                                                    beta: float,
                                                                    payoff_matrix: np.ndarray = None,
                                                                    game: AbstractGame = None,
                                                                    group_size: Optional[int] = 2,
                                                                    figsize: Optional[Tuple[int, int]] = (10, 8),
                                                                    ax: Optional[plt.axis] = None) -> \
        Tuple[Simplex2D, Callable[[np.ndarray, int], np.ndarray], AbstractGame, PairwiseComparison]:
    """
    Helper function to simplify the plotting of the moran dynamics in a 2 Simplex.

    Parameters
    ----------
    population_size:
        Size of the finite population.
    beta:
        Intensity of selection.
    payoff_matrix:
        The square payoff matrix.
    game:
        Game that should contain a set of payoff matrices
    group_size:
        Size of the group. By default, we assume that interactions are pairwise (the group size is 2).
    figsize:
        Size of the figure. This parameter is only used if the ax parameter is not defined.
    ax:
        A matplotlib figure axis.

    Returns
    -------
    A tuple with the simplex object which can be used to add more features to the plot, the function that
    can be used to calculate gradients and should be passed to `Simplex2D.draw_trajectory_from_roots` and
    `Simplex2D.draw_scatter_shadow`, a list of the roots in barycentric coordinates, a list of the roots in
    cartesian coordinates and a list of booleans indicating whether the roots are stable. It also returns the
    game class (this is important, since a new game is created when passing a payoff matrix, and if not returned,
    a reference to the game instance will disappear, and it will produce a segmentation fault). Finally, it also returns
    a reference to the evolver object.
    """
    if (payoff_matrix is None) and (game is None):
        raise Exception("You need to define either a payoff matrix or a game.")
    elif game is None:
        if (group_size is None) or (group_size < 2):
            raise Exception("group_size not be None and must be >= 2")

        if group_size == 2:
            game = Matrix2PlayerGameHolder(payoff_matrix.shape[0], payoff_matrix)
        else:
            game = MatrixNPlayerGameHolder(payoff_matrix.shape[0], group_size, payoff_matrix)

    simplex = Simplex2D(discrete=True, size=population_size, nb_points=population_size + 1)
    simplex.add_axis(figsize, ax)

    v = np.asarray(xy_to_barycentric_coordinates(simplex.X, simplex.Y, simplex.corners))
    v_int = np.floor(v * population_size).astype(np.int64)

    evolver = PairwiseComparison(population_size=population_size, game=game)
    result = np.zeros(shape=(v_int.shape[1], v_int.shape[2], 3))
    for i in range(v_int.shape[1]):
        for j in range(v_int.shape[2]):
            if v_int[:, i, j].sum() <= population_size:
                result[i, j, :] = evolver.calculate_gradient_of_selection(beta, v_int[:, i, j])

    result = result.swapaxes(0, 1).swapaxes(0, 2)
    xy_results = vectorized_barycentric_to_xy_coordinates(result, simplex.corners)

    ux = xy_results[:, :, 0].astype(np.float64)
    uy = xy_results[:, :, 1].astype(np.float64)

    simplex.apply_simplex_boundaries_to_gradients(ux, uy)

    return simplex, lambda u, t: population_size * evolver.calculate_gradient_of_selection(beta, u), game, evolver



import numpy as np

from typing import Union, List
from abc import abstractmethod
from . import AbstractGame, AbstractNPlayerGame
from .. import calculate_state, sample_simplex


class AbstractNPlayerGameExpectedPayoff(AbstractNPlayerGame):
    """
    This abstract Game class can be used in most scenarios where the fitness of a strategy is calculated as its
    expected payoff given the population state.

    It assumes that the game is N player, since the fitness of a strategy given a population state is calculated
    as the expected payoff of that strategy over all possible group combinations in the given state.

    Notes
    -----
    It might be a good idea to overwrite the methods `__str__`, `type`, and `save_payoffs` to adapt to your
    given game implementation

    It assumes that you have at least the following attributes:
     1. And an attribute `self.nb_strategies_` which contains the number of strategies
     that you are going to analyse for the given game.
     2. `self.payoffs()` returns a numpy.ndarray and contain the payoff matrix of the game. This array
     is of shape (self.nb_strategies(), self.nb_group_configurations()), where self.nb_group_configurations()
     is the number
     of possible combinations of strategies in the group. Thus, each row should give the (expected) payoff of the row
     strategy when playing in a group with the column configuration. The `payoff` method provides an easy way to access
     the payoffs for any group composition, by taking as arguments the index of the row strategy
     and a List with the count of each possible strategy in the group.

     You must still implement the methods `play` which should define how the game assigns
     payoffs to each strategy for a given game context. In particular, `calculate_payoffs` should fill the
     array `self.payoffs_` with the correct values as explained above. We recommend that you run this method in
     the `__init__` (initialization of the object) since, these values must be set before passing the game object
     to the numerical simulator (e.g., egttools.numerical.PairwiseComparisonNumerical).

    """

    @abstractmethod
    def play(self, group_composition: Union[List[int], np.ndarray], game_payoffs: np.ndarray) -> None:
        """
        This method fills the `game_payoffs` container with the payoff of each strategy given the `group_composition`.

        Strategies not present in the group will receive 0 payoff by default.

        Parameters
        ----------
        group_composition: Union[List[int], numpy.ndarray]
            A List or a numpy.ndarray containing the counts of each strategy in the group (e.g., for a game with 3
            possible strategies and group size 4, the following List is possible [3, 0, 1]).
        game_payoffs: numpy.ndarray
            A container for the payoffs that will be calculated. This avoids needing to create a new array at each
            call and should speed up computation.
        """
        pass

    def calculate_payoffs(self) -> np.ndarray:
        """
        This method calculates the payoffs for each strategy in each possible group configuration. Thus, it must
        fill the `self.payoffs_` numpy.ndarray with these payoffs values. This array
        must be of shape (self.nb_strategies_, nb_group_configurations), where nb_group_configurations is the number
        of possible combinations of strategies in the group. Thus, each row should give the (expected) payoff of the row
        strategy when playing in a group with the column configuration.

        Returns
        -------
        numpy.ndarray
            The payoff matrix of the game.

        """
        payoffs_container = np.zeros(shape=(self.nb_strategies(),), dtype=np.float64)
        for i in range(self.nb_group_configurations()):
            # Get group composition
            group_composition = sample_simplex(i, self.group_size(), self.nb_strategies())
            self.play(group_composition, payoffs_container)
            for strategy_index, strategy_payoff in enumerate(payoffs_container):
                self.update_payoff(strategy_index, i, strategy_payoff)
            # Reinitialize payoff vector
            payoffs_container[:] = 0

        return self.payoffs()


class AbstractTwoPLayerGame(AbstractGame):
    """
    This abstract Game class can be used in most scenarios where the fitness of a strategy is calculated as its
    expected payoff given the population state.

    It assumes that the game is 2 player and the fitness is calculated with this assumption!

    Notes
    -----
    It might be a good idea to overwrite the methods `__str__`, `type`, and `save_payoffs` to adapt to your
    given game implementation

    It assumes that you have at least the following attributes:
     1. And an attribute `self.nb_strategies_` which contains the number of strategies
     that you are going to analyse for the given game.
     2. `self.payoffs_` which must be a numpy.ndarray and contain the payoff matrix of the game. This array
     must be of shape (self.nb_strategies_, self.nb_strategies_).

    For normal form games:
    1. There is already a class called NormalFormGame available which you can use for these types of games. If
    for any reason this does not cover your needs then:
    2. If your game is normal form, but iterated, you should create another variable to contain the payoff matrix
    for one round of the game, since `self.payoffs_` will contain the expected payoffs over the several rounds
    of the game.
    3. If the game is one-shot and normal form, `self.payoffs_` is the payoff matrix of the game, and you do not
    need to do anything in calculate_payoffs besides calling this matrix.


     You must still implement the methods `play` and `calculate_payoffs` which should define how the game assigns
     payoffs to each strategy for each possible game context. In particular, `calculate_payoffs` should fill the
     array `self.payoffs_` with the correct values as explained above. We recommend that you run this method in
     the `__init__` (initialization of the object) since, these values must be set before passing the game object
     to the numerical simulator (e.g., egttools.numerical.PairwiseComparisonNumerical).

    """

    def __init__(self, nb_strategies: int):
        """
        This class must be initialized with the total number of strategies
        that will be used and the size of the group in which the game takes place.
        This is required to calculate the number of group configurations and the correct
        shape of the payoff matrix.

        Parameters
        ----------
        nb_strategies: int
            total number of possible strategies.
        """
        super().__init__()
        self.nb_strategies_ = nb_strategies
        self.payoffs_ = np.zeros(shape=(self.nb_strategies_, self.nb_strategies_))

        # initialize the payoffs matrix
        self.calculate_payoffs()

    @abstractmethod
    def play(self, group_composition: Union[List[int], np.ndarray], game_payoffs: np.ndarray) -> None:
        """
        This method fills the `game_payoffs` container with the payoff of each strategy given the `group_composition`.

        Strategies not present in the group will receive 0 payoff by default.

        Parameters
        ----------
        group_composition: Union[List[int], numpy.ndarray]
            A List or a numpy.ndarray containing the counts of each strategy in the group (e.g., for a game with 3
            possible strategies and group size 4, the following List is possible [3, 0, 1]).
        game_payoffs: numpy.ndarray
            A container for the payoffs that will be calculated. This avoids needing to create a new array at each
            call and should speed up computation.
        """
        pass

    @abstractmethod
    def calculate_payoffs(self) -> np.ndarray:
        """
        This method calculates the payoffs for each strategy in each possible group configuration. Thus, it must
        fill the `self.payoffs_` numpy.ndarray with these payoffs values. This array
        must be of shape (self.nb_strategies_, nb_group_configurations), where nb_group_configurations is the number
        of possible combinations of strategies in the group. Thus, each row should give the (expected) payoff of the row
        strategy when playing in a group with the column configuration.

        Returns
        -------
        numpy.ndarray
            The payoff matrix of the game.

        """
        pass

    def calculate_fitness(self, player_strategy: int, pop_size: int, population_state: np.ndarray) -> float:
        """
        Calculates the Fitness of a strategy for a given population state.

        The calculation is done by computing the expected payoff over all possible strategy matches.

        Parameters
        ----------
        player_strategy : int
            index of the strategy.
        pop_size : int
            size of the population - Only necessary for compatibility with the C++ implementation
            (might be eliminated in the future).
        population_state : numpy.ndarray[numpy.uint64[m, 1]]
            vector with the population state (the number of players adopting each strategy).

        Returns
        -------
        float
            The fitness of the population.
        """

        population_state[player_strategy] -= 1
        fitness = 0.0
        for i in range(self.nb_strategies_):
            fitness += (population_state[i] / (pop_size - 1)) * self.payoffs_[player_strategy, i]
        population_state[player_strategy] += 1

        return fitness

    def __str__(self) -> str:
        return "AbstractTwoPLayerGame"

    def nb_strategies(self) -> int:
        return self.nb_strategies_

    def type(self) -> str:
        return "AbstractTwoPLayerGame"

    def payoffs(self) -> np.ndarray:
        return self.payoffs_

    def payoff(self, strategy: int, group_composition: List[int]) -> float:
        if strategy > self.nb_strategies_:
            raise IndexError(f'You must specify a valid index for the strategy [0, {self.nb_strategies_}].')
        elif len(group_composition) != self.nb_strategies_:
            raise Exception(f'The group composition list must be of size {self.nb_strategies_}')

        return self.payoffs_[strategy, calculate_state(self.group_size_, group_composition)]

    def save_payoffs(self, file_name: str) -> None:
        with open(file_name, 'w') as f:
            f.write('Payoff matrix of the game:\n')
            f.write(f'{self.payoffs_}')



"""Helpful implementations of stochastic distributions."""

try:
    from ..numerical.numerical_.distributions import (TimingUncertainty,
                                                      multinomial_pmf,
                                                      multivariate_hypergeometric_pdf,
                                                      binom,
                                                      comb)
except Exception:
    raise Exception("numerical package not initialized")

__all__ = ['TimingUncertainty', 'multinomial_pmf', 'multivariate_hypergeometric_pdf', 'binom', 'comb']



import pytest

import os
from sys import platform

import numpy as np
egt = pytest.importorskip("egttools")

Random = egt.Random
PairwiseComparisonNumerical = egt.numerical.PairwiseComparisonNumerical
NormalFormGame = egt.games.NormalFormGame


# os.environ['KMP_STACKSIZE'] = '4m'
# os.environ['KMP_VERSION'] = '.TRUE.'

def test_openmp_simulations():
    # Payoff matrix
    V = 2
    D = 3
    T = 1
    Z = 100
    A = np.array([
        [(V - D) / 2, V],
        [0, (V / 2) - T],
    ])

    # Necessary to avoid issues with Anaconda on MacOSX
    # if platform == "darwin":
    #     os.environ['KMP_DUPLICATE_LIB_OK'] = 'True'

    Random.init(3610063510)

    assert Random._seed == 3610063510

    game = egt.games.NormalFormGame(1, A)
    assert game.type() == 'NormalFormGame'

    evolver = PairwiseComparisonNumerical(Z, game, 10000)
    dist = evolver.estimate_stationary_distribution_sparse(10, 100000, 1000, 1, 1e-3)

    assert dist.toarray().shape == (1, Z + 1)


if __name__ == "__main__":
    test_openmp_simulations()



import pytest

import numpy as np

egttools = pytest.importorskip("egttools")

PGG = egttools.games.PGG
player_factory = egttools.behaviors.pgg_behaviors.player_factory


def test_extend_abstract_game() -> None:
    """
    This test checks that a Normal form game runs as expected
    """
    strategies = player_factory([0, 1])

    game = PGG(group_size=2, cost=np.float64(1.0), multiplying_factor=np.float64(3.0), strategies=strategies)

    payoffs = np.array([
        [0., 1.5, 0.],
        [0., 0.5, 2.]
    ])

    assert game.nb_strategies() == 2
    assert game.type() == "PGG"
    np.testing.assert_array_equal(game.payoffs(), payoffs)
